from __future__ import print_function
from __future__ import print_function
from __future__ import print_function
import unittest

from nose.tools import raises

from decent_params import DecentParams, UserError, DecentParamsUnknownArgs


class ParamsTest(unittest.TestCase):
        
    def decent_params_test0(self):
        
        p = DecentParams()
        p.add_float_list('floats')
        
        args = ['--floats', '1.2', '2.3']
        
        try:    
            res = p.parse_args(args)
        except SystemExit as e:
            print(e)
            raise Exception(str(e))
        
        self.assertEqual(res.given('floats'), True)
        self.assertEqual(res.given('int1'), False)
        self.assertEqual(res['floats'], [1.2, 2.3])
        
    def decent_params_test1(self):
        
        p = DecentParams()
        p.add_string('vehicle', default='x')
        p.add_float('float1')
        p.add_float_list('floats')
        p.add_int('int1', default=0)
        p.add_int_list('ints')
        p.add_string_choice('ciao', ['2', '1'], default='1')
        p.add_int_choice('ciao2', [2, 1], default=1)
        
        args = ['--vehicle', 'y',
                '--float1', '1.2',
                '--floats', '1.2', '2.3',
                '--ints', '1', '2', '3']
        
        try:    
            res = p.parse_args(args)
        except SystemExit as e:
            print(e)
            raise Exception(str(e))
        
        self.assertEqual(res.given('floats'), True)
        self.assertEqual(res.given('int1'), False)
        self.assertEqual(res['ciao'], '1')
        self.assertEqual(res['ciao2'], 1)
        self.assertEqual(res.ciao, '1')
        self.assertEqual(res.ciao2, 1)
        self.assertEqual(res.floats, [1.2, 2.3])
        self.assertEqual(res.ints, [1, 2, 3])
        
    @raises(UserError)
    def decent_params_test2(self):
        """ Test compulsory """
        p = DecentParams()
        p.add_string('a')
        p.add_string('b')
        res = p.parse_args(['--b', '1'])
        print(res)
        
    
    def decent_params_test_notdefined(self):
        p = DecentParams()
        p.add_string('vehicle', default='x')
        
        self.assertRaises(DecentParamsUnknownArgs, p.parse_args, ['--b', '1'])
        



from .utils import UserError

__all__ = [
   'DecentParamsUserError',
   'DecentParamsUnknownArgs',
   'DecentParamsSemanticError',
]


class DecentParamsDefinitionError(Exception):
    """ These are raised when constructing data structures """
    pass
    
class DecentParamsUserError(UserError):
    """ These are raised when interrpreting arguments """
    def __init__(self, dp, msg):
        self.dp = dp
        self.msg = msg
        s = msg
#         s += '\n' + str(dp)
        Exception.__init__(self, s)

class DecentParamsSemanticError(DecentParamsUserError):
    def __init__(self, dp, p, msg):
        self.p = p
        s = msg + '\n while interpreting parameter %s ' % p
        DecentParamsUserError.__init__(self, dp, s)
    
class DecentParamsUnknownArgs(DecentParamsUserError):
    
    def __init__(self, dp, unknown):
        msg = 'Could not interpret %r.' % unknown
        msg += '\n' + 'Known parameters: %s.' % ', '.join(dp.params)
        DecentParamsUserError.__init__(self, dp, msg)



import logging

class HasLogger(object):
    
    def __init__(self, logger=None):
        if logger is None:
            logger = logging.getLogger()
            logger.setLevel(logging.DEBUG)
        self.logger = logger
        
    def info(self, *args, **kwargs):
        return self.logger.info(*args, **kwargs)

    def warn(self, *args, **kwargs):
        return self.logger.warn(*args, **kwargs)

    def error(self, *args, **kwargs):
        return self.logger.error(*args, **kwargs)
    
    def debug(self, *args, **kwargs):
        return self.logger.debug(*args, **kwargs)



import os
import shutil
import sys
import traceback
from abc import abstractmethod
from typing import List

import contracts
from compmake import CommandFailed, StorageFilesystem, read_rc_files
from compmake.context import Context
from compmake.exceptions import ShellExitRequested
from compmake.jobs.uptodate import CacheQueryDB
from contracts import ContractsMeta, contract, indent
from decent_params import DecentParams
from decent_params.utils import UserError, wrap_script_entry_point
from quickapp import QUICKAPP_COMPUTATION_ERROR, logger

from .compmake_context import CompmakeContext, context_get_merge_data
from .exceptions import QuickAppException
from .quick_app_base import QuickAppBase
from .report_manager import _dynreports_create_index

__all__ = [
    'QuickApp',
    'quickapp_main',
]


class QuickApp(QuickAppBase):
    """ Template for an application that uses compmake to define jobs. """

    __metaclass__ = ContractsMeta

    # Interface to be implemented
    @abstractmethod
    def define_jobs_context(self, context):
        """ Define jobs in the current context. """

    @abstractmethod
    def define_options(self, params: DecentParams):
        """ Define options for the application. """

    # Implementation

    def _define_options_compmake(self, params):
        script_name = self.get_prog_name()
        s = script_name
        s = s.replace('.py', '')
        s = s.replace(' ', '_')
        default_output_dir = 'out-%s/' % s

        g = 'Generic arguments for Quickapp'
        # TODO: use  add_help=False to ARgParsre
        # params.add_flag('help', short='-h', help='Shows help message')
        params.add_flag('contracts', help='Activate PyContracts', group=g)
        params.add_flag('profile', help='Use Python Profiler', group=g)
        params.add_flag('compress', help='Compress stored data', group=g)
        params.add_string('output', short='o',
                          help='Output directory',
                          default=default_output_dir, group=g)

        params.add_flag('reset', help='Deletes the output directory', group=g)
        # params.add_flag('compmake', help='Activates compmake caching (if app is such that set_default_reset())', group=g)

        params.add_flag('console', help='Use Compmake console', group=g)

        params.add_string('command', short='c',
                          help="Command to pass to compmake for batch mode",
                          default=None, group=g)

    def define_program_options(self, params):
        self._define_options_compmake(params)
        self.define_options(params)

    def get_qapp_parent(self):
        parent = self.parent
        while parent is not None:
            # logger.info('Checking %s' % parent)
            if isinstance(parent, QuickApp):
                return parent
            parent = parent.parent
        return None

    def go(self):
        # check that if we have a parent who is a quickapp,
        # then use its context
        qapp_parent = self.get_qapp_parent()
        if qapp_parent is not None:
            # self.info('Found parent: %s' % qapp_parent)
            qc = qapp_parent.child_context
            self.define_jobs_context(qc)
            return
        else:
            # self.info('Parent not found')
            pass

        # if False:
        #     import resource
        #     gbs = 5
        #     max_mem = long(gbs * 1000 * 1048576)
        #     resource.setrlimit(resource.RLIMIT_AS, (max_mem, -1))
        #     resource.setrlimit(resource.RLIMIT_DATA, (max_mem, -1))

        options = self.get_options()

        # if self.get_qapp_parent() is None:
        # only do this if somebody didn't do it before
        if not options.contracts:
            msg = ('PyContracts disabled for speed. '
                   'Use --contracts to activate.')
            self.logger.warning(msg)
            contracts.disable_all()

        output_dir = options.output

        if options.reset:
            if os.path.exists(output_dir):
                self.logger.info('Removing output dir %r.' % output_dir)
                try:
                    shutil.rmtree(output_dir)
                except OSError as e:
                    # Directory not empty -- common enough on NFS filesystems
                    # print('errno: %r' % e.errno)
                    if e.errno == 39:
                        pass
                    else:
                        raise

        # Compmake storage for results
        storage = os.path.join(output_dir, 'compmake')
        logger.debug('Creating storage in %s  (compress = %s)' % (storage, options.compress))
        db = StorageFilesystem(storage, compress=options.compress)
        currently_executing = ['root']
        # The original Compmake context
        oc = Context(db=db, currently_executing=currently_executing)
        # Our wrapper
        qc = CompmakeContext(cc=oc,
                             parent=None, qapp=self, job_prefix=None,
                             output_dir=output_dir)
        read_rc_files(oc)

        original = oc.get_comp_prefix()
        self.define_jobs_context(qc)
        oc.comp_prefix(original)

        merged = context_get_merge_data(qc)

        # Only create the index job if we have reports defined
        # or some branched context (which might create reports)
        has_reports = len(qc.get_report_manager().allreports) > 0
        has_branched = qc.has_branched()
        if has_reports or has_branched:
            # self.info('Creating reports')
            oc.comp_dynamic(_dynreports_create_index, merged)
        else:
            pass
            # self.info('Not creating reports.')

        ndefined = len(oc.get_jobs_defined_in_this_session())
        if ndefined == 0:
            # self.comp was never called
            msg = 'No jobs defined.'
            raise ValueError(msg)
        else:
            if options.console:
                oc.compmake_console()
                return 0
            else:
                cq = CacheQueryDB(oc.get_compmake_db())
                targets = cq.all_jobs()
                todo, done, ready = cq.list_todo_targets(targets)

                if not todo and  options.command is None:
                    msg = "Note: there is nothing for me to do. "
                    msg += '\n(Jobs todo: %s done: %s ready: %s)' % (len(todo), len(done), len(ready))
                    msg += """\
This application uses a cache system for the results.
This means that if you call it second time with the same arguments,
 and if you do not change any input, it will not do anything."""
                    self.warn(msg)
                    return 0

                if options.command is None:
                    command = 'make recurse=1'
                else:
                    command = options.command

                try:
                    _ = oc.batch_command(command)
                    # print('qapp: ret0 = %s'  % ret0)
                except CommandFailed:
                    # print('qapp: CommandFailed')
                    ret = QUICKAPP_COMPUTATION_ERROR
                except ShellExitRequested:
                    # print('qapp: ShellExitRequested')
                    ret = 0
                else:
                    # print('qapp: else ret = 0')
                    ret = 0

                return ret

    @contract(args='dict(str:*)|list(str)', extra_dep='list')
    def call_recursive(self, context, child_name, cmd_class, args,
                       extra_dep: List=None,
                       add_outdir=None,
                       add_job_prefix=None,
                       separate_resource_manager=False,
                       separate_report_manager=False,
                       extra_report_keys=None):
        if extra_dep is None:
            extra_dep = []
        instance = cmd_class()
        instance.set_parent(self)
        is_quickapp = isinstance(instance, QuickApp)

        try:
            # we are already in a context; just define jobs
            child_context = context.child(qapp=instance, name=child_name,
                                          extra_dep=extra_dep,
                                          add_outdir=add_outdir,
                                          extra_report_keys=extra_report_keys,
                                          separate_resource_manager=separate_resource_manager,
                                          separate_report_manager=separate_report_manager,
                                          add_job_prefix=add_job_prefix)  # XXX

            if isinstance(args, list):
                instance.set_options_from_args(args)
            elif isinstance(args, dict):
                instance.set_options_from_dict(args)
            else:
                assert False

            if not is_quickapp:
                self.child_context = child_context
                res = instance.go()
            else:
                instance.context = child_context
                res = instance.define_jobs_context(child_context)

            # Add his jobs to our list of jobs
            context._jobs.update(child_context.all_jobs_dict())
            return res

        except Exception as e:
            msg = 'While trying to run  %s\n' % cmd_class.__name__
            msg += 'with arguments = %s\n' % args
            if '_options' in instance.__dict__:
                msg += ' parsed options: %s\n' % instance.get_options()
                msg += ' params: %s\n' % instance.get_options().get_params()
            if isinstance(e, QuickAppException):
                msg += indent(str(e), '> ')
            else:
                msg += indent(traceback.format_exc(), '> ')
            raise QuickAppException(msg)


def quickapp_main(quickapp_class, args=None, sys_exit=True):
    """
        Use like this:

            if __name__ == '__main__':
                quickapp_main(MyQuickApp)


        if sys_exit is True, we call sys.exis(ret), otherwise we return the value.

    """
    instance = quickapp_class()
    if args is None:
        args = sys.argv[1:]

    return wrap_script_entry_point(instance.main, logger,
                                   exceptions_no_traceback=(UserError, QuickAppException),
                                   args=args, sys_exit=sys_exit)



import os
from typing import List, Sequence, Tuple

__all__ = ['_context_names_heuristics']


# @contract(objects='seq[N](str)', returns='tuple(str, list[N](str), str)')
def minimal_names_at_boundaries(objects: List[str], separators=['_', '-']) -> Tuple[str, List[str], str]:
    """
        Converts a list of object IDs to a minimal non-ambiguous list of names.

        In this version, we only care about splitting at boundaries
        defined by separators.

        For example, the names: ::

            test_learn1_fast_10
            test_learn1_slow_10
            test_learn2_faster_10

        is converted to: ::

            learn1_fast
            learn2_slow
            learn2_faster

        Returns prefix, minimal, postfix
    """

    if len(objects) == 1:
        return '', objects, ''

    s0 = separators[0]

    # convert and split to uniform separators

    def convert(x: str) -> str:
        return x
        # for s in separators[1:]:
        #     x = x.replace(s, s0)
        # return x

    objectsu = list(map(convert, objects))
    astokens = [x.split(s0) for x in objectsu]

    def is_valid_prefix(p):
        return all(x.startswith(p) for x in objectsu)

    def is_valid_postfix(p):
        return all(x.endswith(p) for x in objectsu)

    # max number of tokens
    ntokens = max(list(map(len, astokens)))
    prefix = None
    for i in range(ntokens):
        toks = astokens[0][:i]
        p = "".join(t + s0 for t in toks)
        if is_valid_prefix(p):
            prefix = p
        else:
            break
    assert prefix is not None

    postfix = None
    for i in range(ntokens):
        t0 = astokens[0]
        toks = t0[len(t0) - i:]
        x = "".join(s0 + t for t in toks)
        if is_valid_postfix(x):
            postfix = x
        else:
            break

    assert postfix is not None

    n1 = len(prefix)
    n2 = len(postfix)
    # remove it
    minimal = [o[n1:len(o) - n2] for o in objectsu]

    # recreate them to check everything is ok
    objects2 = [prefix + m + postfix for m in minimal]

    assert objects == objects2, (objects, objects2, (prefix, minimal, postfix))
    return prefix, minimal, postfix


def _context_names_heuristics(values: List) -> List[str]:
    # print('name heuristics did not work')

    names = get_descriptive_names(values)
    # get nonambiguous and minimal at _,- boundaries
    _, names, _ = minimal_names_at_boundaries(names)
    names = list(map(good_context_name, names))

    return names


def get_descriptive_names(values: List) -> List[str]:
    x = id_field_heuristics(values)
    if x is not None:
        return x

    x = try_heuristics(values, name_field)
    if x is not None:
        return x

    return list(map(str, values))


def name_field(ob):
    if hasattr(ob, '__name__'):
        return getattr(ob, '__name__')
    else:
        return None


def try_heuristics(objects, fun):
    """
        fun must return either a string or None
    """
    names = []
    for o in objects:
        name = fun(o)
        # print('%s -> %s' % (o, name))
        names.append(name)

        if name is None:
            return None

    if len(names) == len(set(names)):
        return names

    return None


def id_field_heuristics(generated):
    # if they are dicts and there's a field 'id', use that as
    # job_id name, otherwise return None
    # (it uses sanitized names)

    alldicts = all([isinstance(g, dict) for g in generated])
    if not alldicts:
        return None

    # find common fields
    fields = set(generated[0].keys())
    for g in generated:
        fields = fields & set(g.keys())

    # print('all fields: %s' % fields)

    is_id_field = lambda x: x.startswith('id')
    id_fields = list(filter(is_id_field, fields))
    if len(id_fields) != 1:
        # print('there are too many fields')
        return None

    id_field = id_fields[0]
    values = [g[id_field] for g in generated]
    # print('Values of %r field are %s' % (id_field, values))
    if len(set(values)) == len(values):
        final = list(map(good_context_name, values))
        return final

    return None



def good_context_name(id_object: str) -> str:
    """
        Removes strange characters from a string to make it a good
        context name.
    """
    id_object = id_object.replace('-', '')
    id_object = id_object.replace('_', '')
    id_object = id_object.replace(' ', '_')
    return id_object


# @contract(objects='seq[N](str)', returns='tuple(str, list[N](str), str)')
def minimal_names(objects: Sequence[str]) -> Tuple[str, List[str], str]:
    """
        Converts a list of object IDs to a minimal non-ambiguous list of names.

        For example, the names: ::

            test_learn_fast_10
            test_learn_slow_10
            test_learn_faster_10

        is converted to: ::

            fast
            slow
            faster

        Returns prefix, minimal, postfix
    """
    if len(objects) == 1:
        return '', objects, ''

    # find the common prefix
    prefix = os.path.commonprefix(objects)
    # invert strings
    objects_i = [o[::-1] for o in objects]
    # find postfix
    postfix = os.path.commonprefix(objects_i)[::-1]

    # print('prefix: %r post: %r' % (prefix, postfix))
    n1 = len(prefix)
    n2 = len(postfix)
    # remove it
    minimal = [o[n1:len(o) - n2] for o in objects]

    # recreate them to check everything is ok
    objects2 = [prefix + m + postfix for m in minimal]

    # print objects, objects2
    assert objects == objects2, (prefix, minimal, postfix)
    return prefix, minimal, postfix



import argparse
import logging
from argparse import ArgumentError, ArgumentParser, PARSER, RawTextHelpFormatter, REMAINDER
from gettext import gettext as _
from pprint import pformat
from typing import Any, Dict, List, Tuple

from decent_params import Choice
from .decent_param import (DecentParam, DecentParamChoice, DecentParamFlag,
                           DecentParamMultiple, DecentParamsResults)
from .exceptions import (DecentParamsDefinitionError, DecentParamsUnknownArgs,
                         DecentParamsUserError)

__all__ = ['DecentParams']


class DecentParams(object):

    def __init__(self, usage=None, prog=None):
        self.usage = usage
        self.prog = prog
        self.params = {}
        self.accepts_extra = False
        self.accepts_extra_description = None

    def __str__(self):
        return 'DecentParams(%s;extra=%s)' % (pformat(self.params), self.accepts_extra)

    def _add(self, p):
        if p.name in self.params:
            msg = "I already know param %r." % p.name
            raise DecentParamsDefinitionError(msg)
        p.order = len(self.params)
        self.params[p.name] = p

    def accept_extra(self, description=None):
        """ Declares that extra arguments are ok. """
        self.accepts_extra = True
        self.accepts_extra_description = description

    def add_flag(self, name, **args):
        if not 'default' in args:
            args['default'] = False
        self._add(DecentParamFlag(ptype=bool, name=name, **args))

    def add_string(self, name, **args):
        S = str
        self._add(DecentParam(ptype=S, name=name, **args))

    def add_float(self, name, **args):
        self._add(DecentParam(ptype=float, name=name, **args))

    def add_int(self, name, **args):
        self._add(DecentParam(ptype=int, name=name, **args))

    def add_bool(self, name, **args):
        self._add(DecentParam(ptype=bool, name=name, **args))

    def add_string_list(self, name, **args):
        self._add(DecentParamMultiple(ptype=str, name=name, **args))

    def add_float_list(self, name, **args):
        self._add(DecentParamMultiple(ptype=float, name=name, **args))

    def add_int_list(self, name, **args):
        self._add(DecentParamMultiple(ptype=int, name=name, **args))

    def add_int_choice(self, name, choices, **args):
        self._add(DecentParamChoice(name=name, choices=choices, ptype=int, **args))

    def add_float_choice(self, name, choices, **args):
        self._add(DecentParamChoice(name=name, choices=choices, ptype=float, **args))

    def add_string_choice(self, name, choices, **args):
        self._add(DecentParamChoice(name=name, choices=choices, ptype=str, **args))

    def parse_args(self, args, allow_choice=True):  # TODO @UnusedVariable
        parser = self.create_parser()
        values, given = self.parse_using_parser(parser, args)
        # TODO: check no Choice()
        res = DecentParamsResults(values, given, self.params)
        return res

    # @contract(args='list(str)', returns='tuple(dict, list(str))')
    def parse_using_parser(self, parser, args: List[str]) -> Tuple[Dict, List[str]]:
        """
            Returns a dictionary with all values of parameters,
            but possibly some values are Choice() instances.

        """
        try:
            argparse_res, argv = parser.parse_known_args(args)
            if argv:
                msg = 'Extra arguments found: %s' % argv
                raise DecentParamsUnknownArgs(self, msg)
        except SystemExit:
            raise  # XXX
            # raise Exception(e)  # TODO

        values, given = self._interpret_args(argparse_res)
        return values, given

    # @contract(args='list(str)', returns='tuple(dict, list(str), list(str))')
    def parse_using_parser_extra(self, parser, args: List[str]) -> Tuple[Dict, List[str], List[str]]:
        """
            This returns also the extra parameters

            returns: values, given, extra
        """

        try:
            #             if False:
            #                 if self.accepts_extra:
            #                     parser.add_argument('remainder', nargs=argparse.REMAINDER)
            #                 argparse_res, unknown = parser.parse_known_args(args)
            #             else:
            if self.accepts_extra:
                parser.add_argument('remainder', nargs='*', help=self.accepts_extra_description)
            argparse_res, unknown = parser.parse_known_intermixed_args(args)
        #                 print('argparse_res: %s' % argparse_res)
        #                 print('unknown: %s' % unknown)
        except SystemExit:
            raise  # XXX

        if unknown:
            raise DecentParamsUnknownArgs(self, unknown)

        if self.accepts_extra:
            extra = argparse_res.remainder
        else:
            extra = []

        values, given = self._interpret_args(argparse_res)
        # TODO: raise if extra is given

        #         print('extra: %r' % extra)
        return values, given, extra

    def _interpret_args(self, argparse_res):
        parsed = vars(argparse_res)
        return self._interpret_args2(parsed)

    # @contract(parsed='dict', returns='tuple(dict, list(str))')
    def _interpret_args2(self, parsed: Dict) -> Tuple[Dict, List[str]]:
        values = dict()
        given = set()
        for k, v in list(self.params.items()):
            # if v.compulsory and parsed[k] is None:
            if v.compulsory and (not k in parsed or parsed[k] is None):
                msg = 'Compulsory option %r not given.' % k
                raise DecentParamsUserError(self, msg)

            # warnings.warn('Not sure below')
            # if parsed[k] is not None:
            if k in parsed and parsed[k] is not None:
                if parsed[k] != self.params[k].default:
                    given.add(k)
                    if isinstance(self.params[k], DecentParamMultiple):
                        if isinstance(parsed[k], list):
                            values[k] = parsed[k]
                        else:
                            values[k] = [parsed[k]]
                    else:
                        if isinstance(parsed[k], list):
                            if len(parsed[k]) > 1:
                                values[k] = Choice(parsed[k])
                            else:
                                values[k] = parsed[k][0]
                        else:
                            values[k] = parsed[k]
                    # values[k] = self.params[k].value_from_string(parsed[k])
                else:
                    values[k] = self.params[k].default
            else:
                values[k] = self.params[k].default
        return values, list(given)

    def _populate_parser(self, option_container, prog=None):
        groups = set(p.group for p in list(self.params.values()))
        for g in groups:

            if g is None:  # normal arguments:
                title = 'Arguments for %s' % prog
            else:
                title = str(g)
            description = None
            group = option_container.add_argument_group(title=title, description=description)
            g_params = [p for p in list(self.params.values()) if p.group == g]
            for p in g_params:
                p.populate(group)

    def create_parser(self, prog=None, usage=None, epilog=None,
                      description=None):
        usage0 = usage

        def my_formatter(prog):
            return RawTextHelpFormatter(prog=prog,
                                        max_help_position=90, width=None)

        class MyParser(argparse.ArgumentParser):

            def error(self, msg):
                raise DecentParamsUserError(self, msg)

            def format_help(self):
                formatter = self._get_formatter()

                def add_dash(s, char):
                    dash = char * len(s)
                    return '\n'.join((dash, s, dash))

                s = "Documentation for the command '%s'" % self.prog

                formatter.add_text(add_dash(s, '='))
                # first description
                formatter.add_text(self.description)
                formatter.add_text("\n")

                #                 print('usage0: %r' % usage0)
                #                 print('usage: %r' % self.usage)
                #
                if usage0 is not None:
                    formatter.add_text(add_dash('Usage and examples', '-'))

                    formatter.add_text(usage0)

                formatter.add_text(add_dash('Formal description of arguments', '-'))

                formatter.add_text(self.usage)

                # positionals, optionals and user-defined groups
                for action_group in self._action_groups:
                    formatter.start_section(action_group.title)
                    formatter.add_text(action_group.description)
                    formatter.add_arguments(action_group._group_actions)
                    formatter.end_section()

                # epilog
                formatter.add_text(self.epilog)

                # determine help from format above
                return formatter.format_help()

        # give None here
        parser = MyParser(prog=prog, usage=None, epilog=epilog,
                          description=description,
                          formatter_class=my_formatter)
        self._populate_parser(option_container=parser, prog=prog)
        return parser

    # @contract(args='list(str)')
    def get_dpr_from_args(self, args: List[str], prog=None, usage=None, epilog=None,
                          description=None):
        parser = self.create_parser(prog=prog, usage=usage,
                                    epilog=epilog, description=description)

        values, given, extra = self.parse_using_parser_extra(parser, args)
        if extra and not self.accepts_extra:
            msg = 'Found extra arguments not accepted: %s' % extra
            raise DecentParamsUserError(self, msg)
        dpr = DecentParamsResults(values, given, self, extra=extra)
        return dpr

    # @contract(config='dict(str:*)')
    def get_dpr_from_dict(self, config: Dict[str, Any]):
        args = []
        for k, v in list(config.items()):
            args.append('--%s' % k)
            args.append(v)
        return self.get_dpr_from_args(args)


# from warnings import warn
logging.basicConfig(format='%(levelname)s: Intermix: %(message)s')


def warn(message):
    logging.warning(message)


def parse_intermixed_args(self, args=None, namespace=None):
    args, argv = self.parse_known_intermixed_args(args, namespace)
    if argv:
        msg = _('unrecognized arguments: %s')
        self.error(msg % ' '.join(argv))
    return args


def parse_known_intermixed_args(self, args=None, namespace=None, _fallback=None):
    # self - argparse parser
    # args, namespace - as used by parse_known_args
    # _fallback - action to take if it can't handle this parser's arguments
    #      (default raises an error)
    # returns a namespace and list of extras

    # positional can be freely intermixed with optionals
    # optionals are first parsed with all positional arguments deactivated
    # the 'extras' are then parsed
    # positionals 'deactivated' by setting nargs=0

    positionals = self._get_positional_actions()

    a = [action for action in positionals if action.nargs in [PARSER, REMAINDER]]
    if a:
        if _fallback is None:
            a = a[0]
            err = ArgumentError(a, 'parse_intermixed_args: positional arg with nargs=%s' % a.nargs)
            self.error(str(err))
        else:
            return _fallback(args, namespace)

    if [action.dest for group in self._mutually_exclusive_groups
        for action in group._group_actions if action in positionals]:
        if _fallback is None:
            self.error('parse_intermixed_args: positional in mutuallyExclusiveGroup')
        else:
            return _fallback(args, namespace)

    save_usage = self.usage
    try:
        if self.usage is None:
            # capture the full usage for use in error messages
            self.usage = self.format_usage()[7:]
        for action in positionals:
            # deactivate positionals
            action.save_nargs = action.nargs
            action.nargs = 0
        try:
            namespace, remaining_args = self.parse_known_args(args, namespace)
            for action in positionals:
                # remove the empty positional values from namespace
                if hasattr(namespace, action.dest):
                    delattr(namespace, action.dest)
        except SystemExit:
            # warn('error from 1st parse_known_args')
            raise
        finally:
            # restore nargs and usage before exiting
            for action in positionals:
                action.nargs = action.save_nargs
            # self.usage = save_usage
        # logging.info('1st: %s,%s'%(namespace, remaining_args))
        # parse positionals
        # optionals aren't normally required, but just in case, turn that off
        optionals = self._get_optional_actions()
        for action in optionals:
            action.save_required = action.required
            action.required = False
        for group in self._mutually_exclusive_groups:
            group.save_required = group.required
            group.required = False
        try:
            namespace, extras = self.parse_known_args(remaining_args, namespace)
        except SystemExit:
            warn('error from 2nd parse_known_args')
            raise
        finally:
            # restore parser values before exiting
            for action in optionals:
                action.required = action.save_required
            for group in self._mutually_exclusive_groups:
                group.required = group.save_required
    finally:
        self.usage = save_usage
    return namespace, extras


ArgumentParser.parse_intermixed_args = parse_intermixed_args
ArgumentParser.parse_known_intermixed_args = parse_known_intermixed_args




def all_starts_with(values, c):
    return all([x[0] == c for x in values])

def all_end_with(values, c):
    return all([x[-1] == c for x in values])
        
def remove_common_prefix(values):
    while True:
        c = values[0][0]
        if all_starts_with(values, c):
            values = [x[1:] for x in values]
        else:
            break
    return values
            
def remove_common_suffix(values):
    while True:
        c = values[0][-1]
        if all_end_with(values, c):
            values = [x[:-1] for x in values]
        else:
            break
    return values


def get_compact_unique_strings(values):
    """ Returns unique strings for each value """
    v = map(str, list(values))
    
    def sanitize(s):
        s = s.replace('.', '')
        s = s.replace('/', '_')
    
    v = map(sanitize, v)
    
    if isinstance(values[0], str):
        values = remove_common_prefix(values)
        values = remove_common_suffix(values)
    
    return dict(zip(values, v))
    
    



# from contracts import contract
from typing import Sequence

from .minimal_name import _context_names_heuristics

__all__ = [
    'iterate_context_names',
    'iterate_context_pairs',
    'iterate_context_triplets',

    'iterate_context_names_pair',
    'iterate_context_names_triplet',
    'iterate_context_names_quartet',
]


# @contract(it1='seq[>0]', key='str|None')
def iterate_context_names(context, it1: Sequence, key: str = None):
    """
        Creates child contexts with minimal names.

        Yields tuples of (context, s1)
    """
    # make strings
    values = list(it1)
    if len(values) == 0:
        raise ValueError('Empty iterator: %s' % values)

    # remove '-' and '_'
    names = _context_names_heuristics(values)

    # print('Using names: %s' % names)

    for x, name in zip(values, names):
        e_c = context.child(name)
        if key is not None:
            keys = {key: x}
            e_c.add_extra_report_keys(**keys)
        yield e_c, x


# @contract(it1='seq[>0]', it2='seq[>0]', key1='str|None', key2='str|None')
def iterate_context_names_pair(context, it1: Sequence, it2: Sequence, key1: str = None, key2: str = None):
    """
        Yields tuples of (context, s1, s2).
    """
    for cc, x1 in iterate_context_names(context, it1, key=key1):
        for c, x2 in iterate_context_names(cc, it2, key=key2):
            yield c, x1, x2


iterate_context_pairs = iterate_context_names_pair


# @contract(it1='seq[>0]', it2='seq[>0]', it3='seq[>0]',
#           key1='str|None', key2='str|None', key3='str|None')
def iterate_context_names_triplet(context, it1: Sequence, it2: Sequence, it3: Sequence,
                                  key1: str = None, key2: str = None, key3: str = None):
    """
        Yields tuples of (context, s1, s2, s3).
    """
    for c1, x1 in iterate_context_names(context, it1, key=key1):
        for c2, x2 in iterate_context_names(c1, it2, key=key2):
            for c3, x3 in iterate_context_names(c2, it3, key=key3):
                yield c3, x1, x2, x3


iterate_context_triplets = iterate_context_names_triplet


def iterate_context_names_quartet(context, it1, it2, it3, it4):
    """
        Yields tuples of (context, s1, s2, s3, s4).
    """
    for c1, x1 in iterate_context_names(context, it1):
        for c2, x2 in iterate_context_names(c1, it2):
            for c3, x3 in iterate_context_names(c2, it3):
                for c4, x4 in iterate_context_names(c3, it4):
                    yield c4, x1, x2, x3, x4


def iterate_context_names_quintuplet(context, it1, it2, it3, it4, it5):
    """
        Yields tuples of (context, s1, s2, s3, s4).
    """
    for c1, x1 in iterate_context_names(context, it1):
        for c2, x2 in iterate_context_names(c1, it2):
            for c3, x3 in iterate_context_names(c2, it3):
                for c4, x4 in iterate_context_names(c3, it4):
                    for c5, x5 in iterate_context_names(c4, it5):
                        yield c5, x1, x2, x3, x4, x5



import logging
logging.basicConfig()
import platform


def add_coloring_to_emit_ansi(fn):
    # add methods we need to the class
    def new(*args):
        levelno = args[1].levelno
        if(levelno >= 50):
            color = '\x1b[31m'  # red
        elif(levelno >= 40):
            color = '\x1b[31m'  # red
        elif(levelno >= 30):
            color = '\x1b[33m'  # yellow
        elif(levelno >= 20):
            color = '\x1b[32m'  # green 
        elif(levelno >= 10):
            color = '\x1b[35m'  # pink
        else:
            color = '\x1b[0m'  # normal

        args[1].msg = '%s%s\x1b[0m' % (color , args[1].msg)
        return fn(*args)
    return new

if platform.system() != 'Windows':
    logging.StreamHandler.emit = add_coloring_to_emit_ansi(
                                                    logging.StreamHandler.emit)



#!/usr/bin/env python
from quickapp import QuickMultiCmdApp, quickapp_main


class DemoApp(QuickMultiCmdApp):
    cmd = 'dp'
    
    def define_options(self, params):
        params.add_string('config', help='Config Joint')
        params.add_int('param2', help='Second parameter')

    def initial_setup(self):
        options = self.get_options()
        self.info('Loading configuration from %r.' % options.config)
        self.info('My param2 is %r.' % options.param2)
        


class DemoAppCmd1(DemoApp.get_sub()):
    cmd = 'cmd1'
    short = 'First command'
    
    def define_options(self, params):
        params.add_int('param1', help='First parameter', default=1)
        params.add_int('param2', help='Second parameter')
        
    def define_jobs(self):
        options = self.get_options()
        self.info('My param2 is %r.' % options.param2)
        



class DemoAppCmd2(DemoApp.sub):
    cmd = 'cmd2'
    short = 'Second command'
    
    def define_options(self, params):
        params.add_int('param1', help='First parameter', default=1)
        
    def define_jobs(self):
        pass

        

def subcommands_test1():
    args = ['-o', 'quickapp_test1',
            '-c', 'make all', '--param1', '10', '--param2', '42']
    quickapp_main(DemoApp, args=args)
    

if __name__ == '__main__':
    quickapp_main(DemoApp)
    



import itertools
from compmake import comp
from reprep.report_utils import StoreResults


def all_combinations(params, give_choices=False):
    choices = sorted([k for k in params if isinstance(params[k], Choice)])
    values = [list(params[k]) for k in choices]
    others = dict([(a, b) for a, b in params.items() if a not in choices])    
    for x in itertools.product(*values):
        chosen = dict(zip(choices, x))
        res = dict(chosen)
        res.update(others)
        if give_choices:
            yield res, chosen
        else:
            yield res
            

def all_args_combinations(*args, **kwargs):
    for i, a in enumerate(args):
        kwargs[str(i)] = a
    
    for c, chosen in all_combinations(kwargs, True):
        margs = []
        for i in range(len(args)):
            margs.append(c[str(i)])
            del c[str(i)]
                  
        yield tuple(margs), c, chosen
        

def call_comb(function, *args, **kwargs):
    sr = StoreResults()
    for a, b, chosen in all_args_combinations(*args, **kwargs):
        sr[chosen] = function(*a, **b)
    return sr
        
def comp_comb(function, *args, **kwargs):
    sr = StoreResults()    
    for a, b, chosen in all_args_combinations(*args, **kwargs):
        def s(k):
            return ""
            if isinstance(k, int):
                return ""
            else:
                return str(k)
        pid = '-'.join(['%s%s' % (s(k), v) for k, v in chosen.items()])
        job_id = '%s-%s' % (function.__name__, pid)
        b['job_id'] = job_id
        job = comp(function, *a, **b)
        sr[chosen] = job
    return sr



from typing import List

import six

# from contracts import describe_type, describe_value
from decent_params import Choice
from .exceptions import DecentParamsSemanticError

# from optparse import Option


not_given = 'DefaultNotGiven'


class DecentParam:

    def __init__(self, ptype, name, default=not_given, help: str = None,  # @ReservedAssignment
                 short=None, allow_multi=False, group=None):
        compulsory = (default == not_given)
        if compulsory:
            default = None
        self.ptype = ptype
        self.name = name
        self.default = default
        self.desc = help
        self.compulsory = compulsory
        self.short = short
        self.order = None
        self.allow_multi = allow_multi
        self.group = group

        self.params = None  # the DecentParams structure

        if self.default is not None:
            self.validate(self.default)

    def __repr__(self) -> str:
        return 'DecentParam(%r,%r)' % (self.ptype, self.name)

    def validate(self, value):
        self.check_type(value)

    def set_from_string(self, s: str):
        self._value = self.value_from_string(s)

    def value_from_string(self, s: str):
        """ Possibly returns Choice(options) """
        # print('%s %s' % (s, self.ptype))
        sep = ','
        if isinstance(s, six.string_types) and sep in s:
            return Choice([self.value_from_string(x)
                           for x in s.split(sep)])

        if self.ptype in six.string_types:
            return self.ptype(s)
        if self.ptype == float:
            return float(s)
        if self.ptype == int:
            return int(s)  # TODO: check
        if self.ptype == bool:
            res = bool(s)  # TODO: check
            # print('s %s -> %s' % (s, res))
            return res
        msg = 'Unknown type %r' % self.ptype
        raise DecentParamsSemanticError(self.params, self, msg)

    def check_type(self, x):
        expected = self.ptype
        if self.ptype == float:
            expected = (float, int)

        if not isinstance(x, expected):
            msg = ("For param '%s', expected %s, got %s" %
                   (self.name, self.ptype, x))
            raise DecentParamsSemanticError(self.params, self, msg)

    def get_desc(self):
        desc = self.desc
        if self.compulsory:
            desc = '[*required*] %s' % desc
        elif self.default is not None:
            desc = '[default: %8s] %s' % (self.default, desc)
        return desc

    def populate(self, parser):
        option = '--%s' % self.name

        nargs = '+' if self.allow_multi else 1
        other = dict(help=self.get_desc(), default=self.default,
                     nargs=nargs)
        other['type'] = self.ptype
        if self.short is not None:
            parser.add_argument(self._get_short_option(), option, **other)
        else:
            parser.add_argument(option, **other)

    def _get_short_option(self):
        short = self.short
        short = short.replace('-', '')
        option1 = '-%s' % short
        return option1


class DecentParamsResults:
    def __init__(self, values, given, params, extra=None):
        self._values = values
        self._given = given
        self._params = params
        self._extra = extra

        for k, v in list(values.items()):
            self.__dict__[k] = v

    def __repr__(self):
        return 'DecentParamsResults(%r,given=%r,extra=%r)' % (self._values, self._given, self._extra)

    def __str__(self):
        return 'DPR(values=%s;given=%s;extra=%s)' % (self._values, self._given, self._extra)

    def get_extra(self) -> List[str]:
        return self._extra

    def get_params(self):
        """ Returns the DecentParams structure which originated these results. """
        return self._params

    def __getitem__(self, name: str):
        return self._values[name]

    def given(self, name: str) -> bool:
        return name in self._given


# class MyOption(Option):
#
#     ACTIONS = Option.ACTIONS + ("extend",)
#     STORE_ACTIONS = Option.STORE_ACTIONS + ("extend",)
#     TYPED_ACTIONS = Option.TYPED_ACTIONS + ("extend",)
#     ALWAYS_TYPED_ACTIONS = Option.ALWAYS_TYPED_ACTIONS + ("extend",)
#
#     def take_action(self, action, dest, opt, value, values, parser):
#         if action == "extend":
#             lvalue = value.split(",")
#             values.ensure_value(dest, []).extend(lvalue)
#         else:
#             Option.take_action(
#                 self, action, dest, opt, value, values, parser)

class DecentParamMultiple(DecentParam):
    """ Allow multiple values """

    def __init__(self, ptype, name, default=not_given, **args):
        if default is not not_given:
            if not isinstance(default, list):
                default = [default]
        DecentParam.__init__(self, ptype=ptype, name=name, default=default,
                             **args)

    def value_from_string(self, s: str) -> List:
        values = [DecentParam.value_from_string(self, x) for x in s.split(',')]
        return values

    def validate(self, value):
        if not isinstance(value, list):
            msg = "Should be a list, not %r" % value
            raise DecentParamsSemanticError(self.params, self, msg)
        for x in value:
            self.check_type(x)

    def populate(self, parser):
        option = '--%s' % self.name

        other = dict(nargs='+', type=self.ptype,
                     help=self.get_desc(), default=self.default)

        #         other = dict(nargs=1, type=self.ptype, action='append',
        #                             help=self.get_desc(), default=self.default)

        if self.short is not None:
            parser.add_argument(self._get_short_option(), option, **other)
        else:
            parser.add_argument(option, **other)


class DecentParamFlag(DecentParam):
    def populate(self, parser, default=False):
        option = '--%s' % self.name
        other = dict(help=self.desc, default=default, action='store_true')

        if self.short is not None:
            parser.add_argument(self._get_short_option(), option, **other)
        else:
            parser.add_argument(option, **other)


class DecentParamChoice(DecentParam):
    def __init__(self, choices, **args):
        self.choices = choices
        DecentParam.__init__(self, **args)

    def validate(self, value):
        if not value in self.choices:
            msg = 'Not %r in %r' % (value, self.choices)
            raise DecentParamsSemanticError(self.params, self, msg)



from optparse import IndentedHelpFormatter
import optparse
from . import UserError

class LenientOptionParser(optparse.OptionParser):

    def parse_args(self, args):
        self.arguments = list(args)
        return optparse.OptionParser.parse_args(self, args)

    def error(self, msg):
        # msg = '%s: %s' % (self.get_prog_name(), msg)
        msg += ('\nArguments: %s %s' % 
                (self.get_prog_name(), " ".join(self.arguments)))
        raise UserError(msg)


def MyOptionParser(prog, usage):
    formatter = IndentedHelpFormatter(
                 indent_increment=2,
                 max_help_position=80,
                 width=100,
                 short_first=1)

    parser = LenientOptionParser(prog=prog, formatter=formatter, usage=usage)
    parser.disable_interspersed_args()
    return parser


class CmdOptionParser(LenientOptionParser):
    def __init__(self, prog, usage, args):
        self.given = args
        
        formatter = IndentedHelpFormatter(
                     indent_increment=2,
                     max_help_position=80,
                     width=100,
                     short_first=1)
        
        LenientOptionParser.__init__(self, prog=prog,
                                     usage=usage,
                                     formatter=formatter)
        self.disable_interspersed_args()
        
    def parse(self):
        return self.parse_args(self.given)
        
    def parse_options(self):
        """ Returns only the options, checking no spurious args """
        options, args = self.parse()
        if args:
            raise UserError('Spurious arguments: %s' % args)
        return options



__version__ = '1.1'

import sys
import traceback


class UserError(Exception):
    pass


def wrap_script_entry_point(function, logger,
                            exceptions_no_traceback=(UserError,),
                            args=None, sys_exit=True):
    """
        Wraps the main() of a script.
        For Exception: we exit with value 2.
        
        :param exceptions_no_traceback: list of exceptions for which we 
         just print the error, and return 1.        

    """
    
    # logger.info('wrap_script_entry_point')

    ret = wrap_script_entry_point_noexit(function, logger, exceptions_no_traceback, args)
    
    # logger.info('wrap_script_entry_point ret %s' % ret)
    
    if sys_exit:
        sys.stdout.flush()
        sys.stderr.flush()
        sys.exit(ret)
    else:
        return ret

def wrap_script_entry_point_noexit(function, logger,
                            exceptions_no_traceback=(UserError,),
                            args=None):
    """
        Wraps the main() of a script.
         
        :param exceptions_no_traceback: list of exceptions for which we 
         just print the error, and return 1.
        
    """
    if args is None:
        args = sys.argv[1:]
    try:
        ret = function(args)
        if ret is None:
            ret = 0
        return ret
    except exceptions_no_traceback as e:
        # print('no traceback')
        logger.error(e)
        return 1
    except Exception as e:
        # print('normal exception')
        s = traceback.format_exc()
        logger.error(s)
#         sys.stderr.write(s)
#         sys.stderr.write('\n')
#         sys.stderr.flush()
        return 2



#!/usr/bin/env python
from compmake.jobs import all_jobs
from nose.tools import istest
from quickapp import (CompmakeContext, QuickApp, quickapp_main,
                      iterate_context_names)

from .quickappbase import QuickappTest


def f(name):
    print(name)
    return name


def define_jobs(context, id_name):
    assert isinstance(context, CompmakeContext)
    context.comp(f, id_name)


class QuickAppDemoChild(QuickApp):

    def define_options(self, params):
        pass

    def define_jobs_context(self, context):
        names = ['a', 'b', 'c']
        for c, id_name in iterate_context_names(context, names):
            define_jobs(c, id_name)


# XXX
# @istest
class CompappTestDynamic(QuickappTest):

    def compapp_test1(self):
        args = '--contracts -o %s -c make --compress' % self.root0
        ret_found = quickapp_main(QuickAppDemoChild, args.split(), sys_exit=False)
        self.assertEqual(0, ret_found)
        print('jobs in %s: %s' % (self.db, list(all_jobs(self.db))))
        self.assertJobsEqual('all', ['a-f', 'b-f', 'c-f'])


class QuickAppDemoChild2(QuickApp):

    def define_options(self, params):
        pass

    def define_jobs_context(self, context):
        names = ['a', 'b', 'c']
        for c, id_name in iterate_context_names(context, names):
            c.comp_dynamic(define_jobs, id_name)


@istest
class CompappTestDynamic2(QuickappTest):

    def compapp_test1(self):
        self.run_quickapp(qapp=QuickAppDemoChild2, cmd='ls')

        # These are the jobs currently defined
        self.assertJobsEqual('all', ['a-define_jobs',
                                     'b-define_jobs',
                                     'c-define_jobs',
                                     'a-context',
                                     'b-context',
                                     'c-context'])

        self.assert_cmd_success_script('make *-define_jobs; ls')

        self.assertJobsEqual('all', ['a-define_jobs',
                                     'b-define_jobs',
                                     'c-define_jobs',
                                     'a-context',
                                     'b-context',
                                     'c-context',
                                     'a-f', 'b-f', 'c-f'])

        self.assertJobsEqual('done', ['a-define_jobs',
                                      'b-define_jobs',
                                      'c-define_jobs',
                                      'a-context',
                                      'b-context',
                                      'c-context'])

        self.assert_cmd_success_script('make; ls')

        self.assertJobsEqual('done', ['a-define_jobs',
                                      'b-define_jobs',
                                      'c-define_jobs',
                                      'a-context',
                                      'b-context',
                                      'c-context',
                                      'a-f', 'b-f', 'c-f'])



from nose.tools import istest

from quickapp import QuickApp, iterate_context_names
from reprep import Report

from .quickappbase import QuickappTest

def report_example1(param1, param2):
    r = Report()
    r.text('type', 'This is one report')
    r.text('param1', '%s' % param1)
    r.text('param2', '%s' % param2)
    return r

def report_example2(param1, param2):
    r = Report()
    r.text('type', 'This is another report')
    r.text('param1', '%s' % param1)
    r.text('param2', '%s' % param2)
    return r

def instance_reports(context):
    param1s = ['a', 'b']
    param2s = [1, 2]
    for c1, param1 in iterate_context_names(context, param1s):
        c1.add_extra_report_keys(param1=param1)
        for c2, param2 in iterate_context_names(c1, param2s):
            c2.add_extra_report_keys(param2=param2)
            r = c2.comp(report_example1, param1=param1, param2=param2)
            c2.add_report(r, 'report_example1')
            r = c2.comp(report_example2, param1=param1, param2=param2)
            c2.add_report(r, 'report_example2')


class QuickAppDemoReport(QuickApp):

    def define_options(self, params):
        pass

    def define_jobs_context(self, context):
        context.comp_dynamic(instance_reports) 


@istest
class CompappTest1(QuickappTest):

    def compapp_test1(self):
        self.run_quickapp(QuickAppDemoReport, cmd='ls') 
        self.assert_cmd_success('make recurse=1')



__version__ = "6.0.4"

from zuper_commons.logs import ZLogger

logger = ZLogger(__name__)

# error in computation
QUICKAPP_COMPUTATION_ERROR = 2

# error in passing parameters
QUICKAPP_USER_ERROR = 1

from .utils import col_logging

from .quick_app_base import *
from .quick_multi_app import *
from .resource_manager import *
from .report_manager import *
from .quick_app import *
from .compmake_context import *
from .app_utils import *

symbols = [QuickMultiCmdApp, QuickApp,
           QuickAppBase, add_subcommand, ResourceManager]

for s in symbols:
    s.__module__ = 'quickapp'



from nose.tools import istest

from quickapp import QuickApp, iterate_context_names

from .quickappbase import QuickappTest


def f(name):
    print(name)
    return name

def define_jobs2(context, id_name):
    context.comp(f, id_name)

def define_jobs1(context, id_name):
    context.comp_dynamic(define_jobs2, id_name)

class QuickAppDemoChild3(QuickApp):

    def define_options(self, params):
        pass

    def define_jobs_context(self, context):
        names1 = ['a', 'b']
        names2 = ['m', 'n']
        for c1, name1 in iterate_context_names(context, names1):
            for c2, name2 in iterate_context_names(c1, names2):
                c2.comp_dynamic(define_jobs1, name1 + name2)

@istest
class TestDynamic3(QuickappTest):

    howmany = None  # used by cases()

    def test_dynamic1(self):
        self.run_quickapp(qapp=QuickAppDemoChild3, cmd='ls')
        
        self.assert_cmd_success('check_consistency')

        self.assertJobsEqual('all', ['a-m-context',
                                     'a-n-context',
                                     'b-m-context',
                                     'b-n-context',
                                     'a-m-define_jobs1',
                                     'b-m-define_jobs1',
                                     'a-n-define_jobs1',
                                     'b-n-define_jobs1'])
        self.assert_cmd_success('make *-define_jobs1')

        self.assert_cmd_success('ls')
        self.assertJobsEqual('all', ['a-m-context',
                                     'a-n-context',
                                     'b-m-context',
                                     'b-n-context',
                                     'a-m-define_jobs1',
                                     'b-m-define_jobs1',
                                     'a-n-define_jobs1',
                                     'b-n-define_jobs1',
                                     'b-m-context-0',
                                     'a-n-context-0',
                                     'a-m-context-0',
                                     'b-n-context-0',
                                     'a-m-define_jobs2',
                                     'b-m-define_jobs2',
                                     'a-n-define_jobs2',
                                     'b-n-define_jobs2',
                                     ])
        self.assert_cmd_success('make;ls')
        self.assertJobsEqual('all', ['a-m-context',
                                     'a-n-context',
                                     'b-m-context',
                                     'b-n-context',
                                     'a-m-define_jobs1',
                                     'b-m-define_jobs1',
                                     'a-n-define_jobs1',
                                     'b-n-define_jobs1',
                                     'b-m-context-0',
                                     'a-n-context-0',
                                     'a-m-context-0',
                                     'b-n-context-0',
                                     'a-m-define_jobs2',
                                     'b-m-define_jobs2',
                                     'a-n-define_jobs2',
                                     'b-n-define_jobs2',
                                     'a-m-f',
                                     'a-n-f',
                                     'b-m-f',
                                     'b-n-f',
                                     ])
        self.assert_cmd_success('details a-m-f')
        self.assert_defined_by('a-m-f', ['root', 'a-m-define_jobs1', 'a-m-define_jobs2'])



from nose.tools import istest

from quickapp import QuickApp

from .quickappbase import QuickappTest


def f():
    return 1

def define_jobs2(context):
    context.comp(f)

def define_jobs1(context):
    context.comp_dynamic(define_jobs2)
    
class QuickAppDemoChild1(QuickApp):

    def define_options(self, params):
        pass

    def define_jobs_context(self, context):
        context.comp_dynamic(define_jobs1)

@istest
class TestDynamic1(QuickappTest):

    howmany = None  # used by cases()

    def test_dynamic1(self):
        self.run_quickapp(qapp=QuickAppDemoChild1, cmd='ls')
        self.assert_cmd_success('make define_jobs1')
        self.assert_cmd_success('ls')
        
        self.assert_cmd_success('make')
        self.assert_cmd_success('make')
        self.assert_cmd_success('make')
        
        




from compmake.jobs.storage import all_jobs
from compmake.unittests.compmake_test import CompmakeTest
from quickapp import quickapp_main


class QuickappTest(CompmakeTest):
    """ Utilities for quickapp testing """

    def run_quickapp(self, qapp, cmd: str):
        args = ['-o', self.root0, '-c', cmd, '--compress']
        self.assertEqual(0, quickapp_main(qapp, args, sys_exit=False))

        # tell the context that it's all good
        jobs = all_jobs(self.db)
        self.cc.reset_jobs_defined_in_this_session(jobs)



import unittest
from quickapp.app_utils.minimal_name import minimal_names_at_boundaries, minimal_names


class TestMinimal(unittest.TestCase):

    def test_minimal_names_at_boundaries(self):
        objects = ['test_random_dpx1_64_10', 'test_random_drob1_64_10']
        
        prefix, minimal, postfix = minimal_names_at_boundaries(objects)
        
        self.assertEqual(prefix, 'test_random_')
        self.assertEqual(postfix, '_64_10')
        
        self.assertEqual(minimal, ['dpx1', 'drob1'])


    def test_minimal_names(self):
        objects = ['test_random_dpx1_64_10', 'test_random_drob1_64_10']
        
        prefix, minimal, postfix = minimal_names(objects)
        
        self.assertEqual(prefix, 'test_random_d')
        self.assertEqual(postfix, '1_64_10')
        
        self.assertEqual(minimal, ['px', 'rob'])



from . import all_args_combinations, comp_comb, all_combinations, Choice


def all_combinations_test1():

#    samples = [
#        (all_combinations(1, 2, a=3),
#         [(1, 2), dict(a=3)]),
#        (all_combinations(Choice([1, 10]), 2, a=3),
#         [(1, 2), dict(a=3)],
#         [(10, 2), dict(a=3)]),
#    ]
#          
    samples = [
        (
            dict(a=Choice([1, 10]), b=3, c=Choice([3, 4])),
            [{'a': 1, 'c': 3, 'b': 3}, {'a': 1, 'c': 4, 'b': 3},
             {'a': 10, 'c': 3, 'b': 3}, {'a': 10, 'c': 4, 'b': 3}]
        ),
        (
            dict(a=Choice([1, 10]), b=3),
            [dict(a=1, b=3), dict(a=10, b=3)]
        ),
    ]
    
    for a, b in samples:
        
        res = list(all_combinations(a))
        assert res == b
    
    
def all_args_combinations_test1():
    
    for a, b, c in all_args_combinations(1, Choice(['a', 'b']), a=3, b=Choice([1, 2])):
        print a, b, c



def f(p1, p2):
    pass


def comp_comb_test1():
    res = comp_comb(f, Choice([1, 2]), b=Choice([1, 3]))
    print res
    
    



import logging
from abc import abstractmethod
from collections import defaultdict

from conf_tools.utils import indent, termcolor_colored
from decent_params import UserError
from .quick_app_base import QuickAppBase

__all__ = ['QuickMultiCmdApp', 'add_subcommand']


class QuickMultiCmdApp(QuickAppBase):

    def define_program_options(self, params):
        self.define_multicmd_options(params)
        params.accept_extra()

    @abstractmethod
    def define_multicmd_options(self, params):
        pass

    @abstractmethod
    def initial_setup(self):
        pass

    @classmethod
    def get_usage(cls):
        names = cls._get_subs_names()
        commands = ' | '.join(names)
        return '%prog ' + '[--config DIR1:DIR2:...] {%s} [command options]' % commands

    @classmethod
    def get_sub(appcls):
        """ Returns the subclass for the subcommands """
        # mainly because eclipse does not see ".sub" as valid.
        if not hasattr(appcls, 'sub'):
            from contracts import ContractsMeta
            class Register(ContractsMeta):
                def __init__(cls, clsname, bases, clsdict):  # @UnusedVariable @NoSelf
                    ContractsMeta.__init__(cls, clsname, bases, clsdict)
                    if not 'cmd' in clsdict:
                        # print('skpping %r' % cls)
                        return
                    # print('Automatically registering %s>%s' % (cls, clsname))
                    if clsname in ['SubCmd']:
                        return
                    cmds = QuickMultiCmdApp.subs[appcls]
                    cmds.append(cls)

            class SubCmd(QuickAppBase):
                __metaclass__ = Register

            appcls.sub = SubCmd
        return appcls.sub

    def go(self):
        self.initial_setup()

        cmds = self._get_subs_as_dict()
        if not cmds:
            msg = 'No commands defined.'
            raise ValueError(msg)

        extra = self.get_options().get_extra()
        if not extra:
            msg = 'Please specify as a command one of %s.' % self._get_subs_names_fmt()
            raise UserError(msg)  # XXX: use better exception

        cmd_name = extra[0]
        cmd_args = extra[1:]
        if not cmd_name in cmds:
            msg = ('Could not find command %s; please use one of %s.' %
                   (cmd_name, self._get_subs_names_fmt()))
            raise UserError(msg)

        sub = cmds[cmd_name]

        sub_inst = sub()
        assert isinstance(sub_inst, QuickAppBase)
        logger_name = '%s-%s' % (self.get_prog_name(), cmd_name)
        logger = logging.getLogger(logger_name)
        logger.setLevel(logging.DEBUG)
        sub_inst.logger = logger

        sub_inst.main(args=cmd_args, parent=self)

    @classmethod
    def get_epilog(cls):
        subs = cls._get_subs()
        if not subs:
            return 'Warning: no commands defined.'

        s = "Commands: \n"
        s += cls.get_epilog_commands()
        return s

    @classmethod
    def get_epilog_commands(cls):
        s = ""

        for sub in cls._get_subs():
            cmd_name = sub.cmd
            # XXX: fixme
            cmd_short = sub.get_short_description()

            cmd_name = termcolor_colored(cmd_name, attrs=['bold'])
            s += "  %30s  %s\n" % (cmd_name, cmd_short)
            if issubclass(sub, QuickMultiCmdApp):
                s += '\n'
                s += indent(sub.get_epilog_commands(), ' ' * 7)
                s += '\n'

        return s

    @classmethod
    def _get_subs_names_fmt(cls):
        """ Returns 'cmd1, cmd2, cmd3; """
        names = cls._get_subs_names()
        possibilities = ', '.join('%r' % x for x in names)
        return possibilities

    @classmethod
    def _get_subs_as_dict(cls):
        """ Returns a dict: cmd_name -> cmd """
        return dict([(x.cmd, x) for x in cls._get_subs()])

    @classmethod
    def _get_subs_names(cls):
        return [x.cmd for x in cls._get_subs()]

    @classmethod
    def _get_subs(cls):
        return QuickMultiCmdApp.subs[cls]

    # QuickMultiCmdApp subclass -> (list of  QuickAppBase)
    subs = defaultdict(list)


# @deprecated
def add_subcommand(app, cmd):
    # logger.info('Adding command  %s - %s' % (app.cmd, cmd.cmd))
    # cmd_name = cmd.cmd
    cmds = QuickMultiCmdApp.subs[app]
    #     if cmd_name in cmds:
    #         msg = 'Duplicate sub command %r.' % cmd_name
    #         raise ValueError(msg)
    #     cmds[cmd_name] = cmd
    cmds.append(cmd)



from collections import defaultdict
import traceback

# from contracts import contract, describe_type

from compmake import Promise
from conf_tools.utils import check_is_in, indent



__all__ = ['ResourceManager']


class ResourceManager:
    class CannotProvide(Exception):
        pass

    def __init__(self, context):
        from quickapp.compmake_context import CompmakeContext
        assert isinstance(context, CompmakeContext), context
        from reprep.report_utils import StoreResults
        self.allresources = StoreResults()
        self.providers = defaultdict(list)  # rtype => list of providers
        self.make_prefix = {}  # rtype => function to make prefix
        self._context = context

    # @contract(rtype='str')
    def set_resource_provider(self, rtype: str, provider):
        """
            provider: any callable. It will be called with "context" as first
                argument, and with any remaining params. It needs to return
                a Compmake Promise() object (i.e. the output of comp()).

        """
        self.providers[rtype].append(provider)

    def set_resource_prefix_function(self, rtype, make_prefix):
        """
            make_prefix: a function that takes (rtype, **params) and
            returns a string.
        """
        self.make_prefix[rtype] = make_prefix

    # @contract(rtype='str')
    def get_resource(self, rtype: str, **params):
        return self.get_resource_job(self._context, rtype, **params)

    # @contract(rtype='str')
    def get_resource_job(self, context, rtype: str, **params):
        # print('RM %s %s get_resource %s %s' % (id(self), self._context, rtype, params))
        key = dict(rtype=rtype, **params)
        already_done = key in self.allresources
        if already_done:
            return self.allresources[key]

        check_is_in('resource type', rtype, self.providers)

        prefix = self._make_prefix(rtype, **params)
        #         print('adding job prefix %r' % prefix)
        c = context.child(name=rtype, add_job_prefix=prefix, add_outdir=rtype)
        c._job_prefix = prefix
        # Add this point we should check if we already created the job
        ok = []
        errors = []
        for provider in self.providers[rtype]:
            try:
                res_i = provider(c, **params)
                ok.append((provider, res_i))
            except ResourceManager.CannotProvide as e:
                errors.append(e)
            except Exception as e:
                msg = 'Error while trying to get resource.\n'
                msg += ' type: %r params: %s\n' % (rtype, params)
                msg += 'While calling provider %r:\n' % provider
                msg += indent(traceback.format_exc(), '> ')
                raise Exception(msg)

        if not ok:
            msg = 'No provider could create this resource:\n'
            msg += "\n".join('- %s' % str(e) for e in errors)
            raise Exception(msg)

        if len(ok) >= 2:
            msg = 'The same resource could be created by two providers.'
            msg += '\n%s %s' % (rtype, params)
            for prov, _ in ok:
                msg += '\n - %s' % prov
            raise Exception(msg)

        assert len(ok) == 1
        res = ok[0][1]
        self.set_resource(res, rtype, **params)
        return res

    def _make_prefix(self, rtype, **params):
        """ Creates the job prefix for the given resource. """
        # use the user-defined if available
        if rtype in self.make_prefix:
            f = self.make_prefix[rtype]
            return f(rtype, **params)

        keys = sorted(list(params.keys()))

        from quickapp.app_utils.minimal_name import good_context_name
        vals = [good_context_name(str(params[k])) for k in keys]
        rtype = rtype.replace('-', '_')
        alls = [rtype] + vals
        prefix = "-".join(alls)
        return prefix

    # @contract(rtype='str')
    def set_resource(self, goal, rtype: str, **params):
        key = dict(rtype=rtype, **params)
        if not isinstance(goal, Promise):
            msg = 'Warning, resource did not return a Compmake Promise.'
            msg += '\n  key: %s' % key
            msg += '\n type: %s' % type(goal)
            # logger.error(msg)
            raise ValueError(msg)

        self.allresources[key] = goal




import logging
import os
import sys
import traceback
from abc import abstractmethod
from pprint import pformat
from typing import Any, Dict, List, Optional

# from contracts import contract, ContractsMeta, describe_value
from decent_params import DecentParams, DecentParamsResults, DecentParamsUserError, UserError
from quickapp import logger
from zuper_commons.text import indent
from .exceptions import QuickAppException
from .utils import HasLogger

__all__ = [
    'QuickAppBase',
]


class QuickAppBase(HasLogger):
    """
        class attributes used:

            cmd
            usage
            description (deprecated) => use docstring

    """
    # __metaclass__ = ContractsMeta
    options: DecentParamsResults

    def __init__(self, parent=None):
        HasLogger.__init__(self)
        self.parent = parent

        self._init_logger()

    def _init_logger(self):
        logger_name = self.get_prog_name()
        logger = logging.getLogger(logger_name)
        logger.setLevel(logging.DEBUG)
        self.logger = logger

    def __getstate__(self):
        d = dict(self.__dict__)
        del d['logger']
        return d

    def __setstate__(self, d):
        self.__dict__.update(d)
        self._init_logger()

    @abstractmethod
    def define_program_options(self, params):
        """ Must be implemented by the subclass. """
        pass

    @abstractmethod
    def go(self) -> Optional[int]:
        """
            Must be implemented. This should return either None to mean success,
            or an integer error code.
        """

    @classmethod
    def get_sys_main(cls):
        """ Returns a function to be used as main function for a script. """
        from quickapp import quickapp_main

        def entry(args=None, sys_exit=True):
            return quickapp_main(cls, args=args, sys_exit=sys_exit)

        return entry

    @classmethod
    def __call__(cls, *args, **kwargs):
        main = cls.get_sys_main()
        return main(*args, **kwargs)

    @classmethod
    def get_program_description(cls):
        """
            Returns a description for the program. This is by default
            looked in the docstring or in the "description" attribute
            (deprecated).
        """
        if cls.__doc__ is not None:
            # use docstring
            docs = cls.__doc__
        else:
            docs = cls.__dict__.get('description', None)

        if docs is None:
            logger.warn('No description at all for %s' % cls)
        else:
            docs = docs.strip()
        return docs

    @classmethod
    def get_short_description(cls):
        longdesc = cls.get_program_description()
        if longdesc is None:
            return None
        return longdesc.strip()  # Todo: extract first sentence

    @classmethod
    def get_usage(cls):
        """
            Returns an usage string for the program. The pattern ``%prog``
            will be substituted with the name of the program.
        """
        usage = cls.__dict__.get('usage', None)
        if usage is None:
            pass
        else:
            usage = usage.strip()
        return usage

    @classmethod
    def get_epilog(cls):
        """
            Returns the string used as an epilog in the help text.
        """
        return None

    @classmethod
    def get_prog_name(cls):
        """
            Returns the string used as the program name. By default
            it is contained in the ``cmd`` attribute.
        """
        if not 'cmd' in cls.__dict__:
            return os.path.basename(sys.argv[0])
        else:
            return cls.__dict__['cmd']

    def get_options(self):
        return self.options

    def set_parent(self, parent):
        self.parent = parent

    def get_parent(self):
        if self.parent is not None:
            assert self.parent != self
        return self.parent

    # @contract(args='None|list(str)', returns=int)
    def main(self, args: Optional[List[str]]=None, parent=None) -> int:
        """ Main entry point. Returns an integer as an error code. """

        if "short" in type(self).__dict__:
            msg = 'Class %s uses deprecated attribute "short".' % type(self)
            msg += ' Use "description" instead.'
            self.error(msg)

        # Create the parameters and set them using args
        self.parent = parent
        self.set_options_from_args(args)

        profile = os.environ.get('QUICKAPP_PROFILE', False)

        if not profile:
            ret = self.go()
        else:

            import cProfile
            out = profile
            print('writing to %r' % out)
            ret = cProfile.runctx('self.go()', globals(), locals(), out)
            import pstats

            p = pstats.Stats(out)
            n = 30
            p.sort_stats('cumulative').print_stats(n)
            p.sort_stats('time').print_stats(n)

        if ret is None:
            ret = 0

        if isinstance(ret, int):
            return ret
        else:
            msg = f'Expected None or an integer fomr self.go(), got {ret}'
            raise ValueError(msg)

    def set_options_from_dict(self, config: Dict[str, Any]):
        """
            Reads the configuration from a dictionary.

            raises: UserError: Wrong configuration, user's mistake.
                    Exception: all other exceptions
        """
        params = DecentParams()
        self.define_program_options(params)

        try:
            self.options = params.get_dpr_from_dict(config)
        except DecentParamsUserError as e:
            raise QuickAppException(str(e))
        except Exception as e:
            msg = 'Could not interpret:\n'
            msg += indent(pformat(config), '| ')
            msg += 'according to params spec:\n'
            msg += indent(str(params), '| ') + '\n'
            msg += 'Error is:\n'
            #             if isinstance(e, DecentParamsUserError):
            #                 msg += indent(str(e), '> ')
            #             else:
            msg += indent(traceback.format_exc(), '> ')
            raise QuickAppException(msg)  # XXX class

    def set_options_from_args(self, args: List[str]):
        """
            Reads the configuration from command line arguments.

            raises: UserError: Wrong configuration, user's mistake.
                    Exception: all other exceptions
        """
        cls = type(self)
        prog = cls.get_prog_name()
        params = DecentParams()
        self.define_program_options(params)

        try:
            usage = cls.get_usage()
            if usage:
                usage = usage.replace('%prog', prog)

            desc = cls.get_program_description()
            epilog = cls.get_epilog()
            self.options = \
                params.get_dpr_from_args(prog=prog, args=args, usage=usage,
                                         description=desc, epilog=epilog)
        except UserError:
            raise
        except Exception as e:
            msg = 'Could not interpret:\n'
            msg += ' args = %s\n' % args
            msg += 'according to params spec:\n'
            msg += indent(str(params), '| ') + '\n'
            msg += 'Error is:\n'
            msg += indent(traceback.format_exc(), '> ')
            raise Exception(msg)  # XXX class



import os
from typing import List, TypeVar, Callable

import six

from compmake import Context, Promise
from compmake.context import load_static_storage
from conf_tools import GlobalConfig
from contracts import  contract, describe_type
from contracts.utils import raise_wrapped
from .report_manager import ReportManager
from .resource_manager import ResourceManager
from zuper_commons.types import check_isinstance

__all__ = [
    'CompmakeContext',
    'QuickAppContext',
]


class QuickAppContext:

    def __init__(self, cc: Context, qapp, parent, job_prefix,
                 output_dir, extra_dep: List = None, resource_manager=None,
                 extra_report_keys=None,
                 report_manager=None):
        check_isinstance(cc, Context)
        check_isinstance(parent, (CompmakeContext, type(None)))
        if extra_dep is None:
            extra_dep = []
        self.cc = cc
        # can be removed once subtask() is removed
        self._qapp = qapp
        # only used for count invocation
        self._parent = parent
        self._job_prefix = job_prefix

        if resource_manager is None:
            resource_manager = ResourceManager(self)

        if report_manager is None:
            self.private_report_manager = True  # only create indexe if this is true
            reports = os.path.join(output_dir, 'report')
            reports_index = os.path.join(output_dir, 'report.html')
            report_manager = ReportManager(self, reports, reports_index)
        else:
            self.private_report_manager = False

        self._report_manager = report_manager
        self._resource_manager = resource_manager
        self._output_dir = output_dir
        self.n_comp_invocations = 0
        self._extra_dep = extra_dep
        self._jobs = {}
        if extra_report_keys is None:
            extra_report_keys = {}
        self.extra_report_keys = extra_report_keys

        self._promise = None

        self.branched_contexts = []
        self.branched_children = []

    def __str__(self)->str:
        return 'CompmakeContext(%s)' % (self._job_prefix)

    # def all_jobs(self):
    #     return list(self._jobs.values())

    def all_jobs_dict(self):
        return dict(self._jobs)

    def checkpoint(self, job_name: str) -> Promise:
        """

            (DEPRECATED)

            Creates a dummy job called "job_name" that depends on all jobs
            previously defined; further, this new job is put into _extra_dep.
            This means that all successive jobs will require that the previous
            ones be done.

            Returns the checkpoint job (CompmakePromise).
        """
        job_checkpoint = self.comp(checkpoint, job_name,
                                   prev_jobs=list(self._jobs.values()),
                                   job_id=job_name)
        self._extra_dep.append(job_checkpoint)
        return job_checkpoint

    #
    # Wrappers form Compmake's "comp".
    #
    def comp(self, f, *args, **kwargs) -> Promise:
        """
            Simple wrapper for Compmake's comp function.
            Use this instead of "comp".
        """
        self.count_comp_invocations()
        self.cc.comp_prefix(self._job_prefix)

        other_extra = kwargs.get('extra_dep', [])
        if isinstance(other_extra, Promise):
            other_extra = [other_extra]

        extra_dep = self._extra_dep + other_extra
        kwargs['extra_dep'] = extra_dep
        promise = self.cc.comp(f, *args, **kwargs)
        self._jobs[promise.job_id] = promise
        return promise

    def comp_dynamic(self, f, *args, **kwargs) -> Promise:
        # XXX: we really dont need it
        context = self._get_promise()
        # context = self

        compmake_args = {}
        compmake_args_name = ['job_id', 'extra_dep', 'command_name']
        for n in compmake_args_name:
            if n in kwargs:
                compmake_args[n] = kwargs[n]
                del kwargs[n]

        if not 'command_name' in compmake_args:
            compmake_args['command_name'] = f.__name__
        #:arg:job_id:   sets the job id (respects job_prefix)
        #:arg:extra_dep: extra dependencies (not passed as arguments)
        #:arg:command_name: used to define job name if job_id not provided.

        both = self.cc.comp_dynamic(_dynreports_wrap_dynamic, qc=context,
                                    function=f, args=args, kw=kwargs,
                                    **compmake_args)

        result = self.comp(_dynreports_getres, both)
        data = self.comp(_dynreports_getbra, both)
        self.branched_contexts.append(data)
        return result

    def comp_config(self, f, *args, **kwargs) -> Promise:
        """
            Like comp, but we also automatically save the GlobalConfig state.
        """
        config_state = GlobalConfig.get_state()
        # so that compmake can use a good name
        if not 'command_name' in kwargs:
            kwargs['command_name'] = f.__name__
        return self.comp(wrap_state, config_state, f, *args, **kwargs)

    def comp_config_dynamic(self, f, *args, **kwargs) -> Promise:
        """ Defines jobs that will take a "context" argument to define
            more jobs. """
        config_state = GlobalConfig.get_state()
        # so that compmake can use a good name
        if not 'command_name' in kwargs:
            kwargs['command_name'] = f.__name__
        return self.comp_dynamic(wrap_state_dynamic, config_state, f, *args, **kwargs)

    def count_comp_invocations(self)->None:
        self.n_comp_invocations += 1
        if self._parent is not None:
            self._parent.count_comp_invocations()

    def get_output_dir(self):
        """ Returns a suitable output directory for data files """
        # only create output dir on demand
        if not os.path.exists(self._output_dir):
            os.makedirs(self._output_dir)

        return self._output_dir

    def child(self, name:str, qapp=None,
              add_job_prefix=None,
              add_outdir=None,
              extra_dep: list=None,
              extra_report_keys=None,
              separate_resource_manager=False,
              separate_report_manager=False) -> "QuickAppContext":
        if extra_dep is None:
            extra_dep = []
        """
            Returns child context

            add_job_prefix =
                None (default) => use "name"
                 '' => do not add to the prefix

            add_outdir:
                None (default) => use "name"
                 '' => do not add outdir

            separate_resource_manager: If True, create a child of the ResourceManager,
            otherwise we just use the current one and its context.
        """
        check_isinstance(name, six.string_types)
        if qapp is None:
            qapp = self._qapp

        name_friendly = name.replace('-', '_')

        if add_job_prefix is None:
            add_job_prefix = name_friendly

        if add_outdir is None:
            add_outdir = name_friendly

        if add_job_prefix != '':
            if self._job_prefix is None:
                job_prefix = add_job_prefix
            else:
                job_prefix = self._job_prefix + '-' + add_job_prefix
        else:
            job_prefix = self._job_prefix

        if add_outdir != '':
            output_dir = os.path.join(self._output_dir, name)
        else:
            output_dir = self._output_dir

        if separate_report_manager:
            if add_outdir == '':
                msg = ('Asked for separate report manager, but without changing output dir. '
                       'This will make the report overwrite each other.')
                raise ValueError(msg)

            report_manager = None
        else:
            report_manager = self._report_manager

        if separate_resource_manager:
            resource_manager = None  # CompmakeContext will create its own
        else:
            resource_manager = self._resource_manager

        _extra_dep = self._extra_dep + extra_dep

        extra_report_keys_ = {}
        extra_report_keys_.update(self.extra_report_keys)
        if extra_report_keys is not None:
            extra_report_keys_.update(extra_report_keys)

        c1 = CompmakeContext(cc=self.cc,
                             qapp=qapp, parent=self,
                             job_prefix=job_prefix,
                             report_manager=report_manager,
                             resource_manager=resource_manager,
                             extra_report_keys=extra_report_keys_,
                             output_dir=output_dir,
                             extra_dep=_extra_dep)
        self.branched_children.append(c1)
        return c1

    @contract(job_id=str)
    def add_job_defined_in_this_session(self, job_id):
        self.cc.add_job_defined_in_this_session(self, job_id)
        if self._parent is not None:
            self._parent.add_job_defined_in_this_session(job_id)

    @contract(extra_dep='list')
    def subtask(self, task, extra_dep=[], add_job_prefix=None, add_outdir=None,
                separate_resource_manager=False,
                separate_report_manager=False,
                extra_report_keys=None,
                **task_config):
        return self._qapp.call_recursive(context=self, child_name=task.cmd,
                                         cmd_class=task, args=task_config,
                                         extra_dep=extra_dep,
                                         add_outdir=add_outdir,
                                         add_job_prefix=add_job_prefix,
                                         extra_report_keys=extra_report_keys,
                                         separate_report_manager=separate_report_manager,
                                         separate_resource_manager=separate_resource_manager)

    # Resource managers
    @contract(returns=ResourceManager)
    def get_resource_manager(self):
        return self._resource_manager

    def needs(self, rtype, **params):
        rm = self.get_resource_manager()
        res = rm.get_resource_job(self, rtype, **params)
        assert isinstance(res, Promise), describe_type(res)
        self._extra_dep.append(res)

    def get_resource(self, rtype, **params):
        rm = self.get_resource_manager()
        return rm.get_resource_job(self, rtype, **params)

    @contract(report=Promise, report_type='str')
    def add_report(self, report, report_type, **params):
        rm = self.get_report_manager()
        params.update(self.extra_report_keys)
        rm.add(self, report, report_type, **params)

    @contract(returns=Promise, report_type='str')
    def get_report(self, report_type, **params):
        """ Returns the promise to the given report """
        rm = self.get_report_manager()
        return rm.get(report_type, **params)

    def get_report_manager(self):
        return self._report_manager

    def add_extra_report_keys(self, **keys):
        for k in keys:
            if k in self.extra_report_keys:
                msg = 'key %r already in %s' % (k, list(self.extra_report_keys))
                raise ValueError(msg)
        self.extra_report_keys.update(keys)

    @contract(returns=Promise)
    def _get_promise(self):
        """ Returns the promise object representing this context. """
        if self._promise is None:
            # warnings.warn('Need IDs for contexts, using job_prefix.')
            # warnings.warn('XXX: Note that this sometimes creates a context '
            #              'with depth 1; then "delete not root" deletes it.')
            self._promise_job_id = 'context'
            self._promise = self.comp(load_static_storage, self,
                                      job_id=self._promise_job_id)
        return self._promise

    def has_branched(self):
        """ Returns True if any comp_dynamic was issued. """
        return (len(self.branched_contexts) > 0 or
                any([c.has_branched() for c in self.branched_children]))


def wrap_state(config_state, f, *args, **kwargs):
    """ Used internally by comp_config() """
    config_state.restore()
    return f(*args, **kwargs)


def wrap_state_dynamic(context, config_state, f, *args, **kwargs):
    """ Used internally by comp_config_dynamic() """
    config_state.restore()
    return f(context, *args, **kwargs)


def checkpoint(name, prev_jobs):
    pass


@contract(context=Context, returns='dict')
def _dynreports_wrap_dynamic(context, qc, function, args, kw):
    """

    """

    qc.cc = context

    res = {}
    try:
        res['f-result'] = function(qc, *args, **kw)
    except TypeError as e:
        msg = 'Could not call %r' % function
        raise_wrapped(TypeError, e, msg, args=args, kw=kw)

    res['context-res'] = context_get_merge_data(qc)
    return res


@contract(branched='list(dict)')
def _dynreports_merge(branched):
    rm = None
    for i, b in enumerate(branched):
        if i == 0:
            rm = b['report_manager']
        else:
            rm.merge(b['report_manager'])
    return dict(report_manager=rm)


@contract(res='dict')
def _dynreports_getres(res):
    """ gets only the result """
    return res['f-result']


@contract(res='dict')
def _dynreports_getbra(res):
    """ gets only the result """
    return res['context-res']


def get_branched_contexts(context):
    """ Returns all promises created by context_comp_dynamic() for this and children. """
    res = list(context.branched_contexts)
    for c in context.branched_children:
        res.extend(get_branched_contexts(c))
    return res


def context_get_merge_data(context):
    rm = context.get_report_manager()
    data = [dict(report_manager=rm)]

    data.extend(get_branched_contexts(context))

    if len(data) > 1:
        return context.cc.comp(_dynreports_merge, data)
    else:
        return data[0]


CompmakeContext = QuickAppContext



from nose.tools import istest

from quickapp import QuickApp, iterate_context_names

from .quickappbase import QuickappTest


def f(name):
    print(name)
    return name

def define_jobs2(context, id_name):
    context.comp(f, id_name)

def define_jobs1(context, id_name):
    context.comp_dynamic(define_jobs2, id_name)

class QuickAppDemoChild3(QuickApp):

    def define_options(self, params):
        pass

    def define_jobs_context(self, context):
        names = ['a', 'b']
        for c, id_name in iterate_context_names(context, names):
            c.comp_dynamic(define_jobs1, id_name)

@istest
class TestDynamic2(QuickappTest):

    howmany = None  # used by cases()

    def test_dynamic1(self):
        self.run_quickapp(qapp=QuickAppDemoChild3, cmd='ls')
        print('ls 1')
        self.assert_cmd_success('ls not *dynrep*')
        self.assertJobsEqual('all', ['a-context',
                                     'b-context',
                                     'a-define_jobs1',
                                     'b-define_jobs1'])
        print('make root') 
        self.assert_cmd_success('make a-define_jobs1 b-define_jobs1')
        print('ls 2')
        self.assert_cmd_success('ls not *dynrep*')
        
        self.assertJobsEqual('all', ['a-context',
                                     'b-context',
                                     'a-define_jobs1',
                                     'b-define_jobs1',
                                     'a-context-0',
                                     'b-context-0',
                                     'a-define_jobs2',
                                     'b-define_jobs2',
                                     ])
        print('make level1')
        self.assert_cmd_success('make level1 level2')
        print('ls 3')
        self.assert_cmd_success('ls not *dynrep*')

        self.assertJobsEqual('all', ['a-context',
                                     'b-context',
                                     'a-define_jobs1',
                                     'b-define_jobs1',
                                     'a-context-0',
                                     'b-context-0',
                                     'a-define_jobs2',
                                     'b-define_jobs2',
                                     'a-f',
                                     'b-f'
                                     ])
        self.assert_cmd_success('details a-f')





#!/usr/bin/env python
from shutil import rmtree
from tempfile import mkdtemp
import os
from nose.tools import istest
from quickapp import (QuickApp, QUICKAPP_USER_ERROR, QUICKAPP_COMPUTATION_ERROR,
                      quickapp_main)
from reprep import Report

from .quickappbase import QuickappTest


def actual_computation(param1, param2):
    print('computing (%s %s)' % (param1, param2))
    return [1, 2, 3, 4]


def report_example(param2, samples):
    print('report_example(%s, %s)' % (param2, samples))
    if param2 == -1:
        print('generating exception')
        raise Exception('fake exception')
    r = Report()
    r.text('samples', str(samples))
    print('creating report')
    return r


class QuickAppDemo2(QuickApp):
    cmd = 'quick-app-demo'

    def define_options(self, params):
        params.add_int('param1', help='First parameter', default=1)
        params.add_int('param2', help='Second parameter')

    def define_jobs_context(self, context):
        options = self.get_options()
        param1 = options.param1
        param2 = options.param2
        samples = context.comp(actual_computation, param1=param1, param2=param2)

        rj = context.comp(report_example, param2, samples)
        context.add_report(rj, 'report_example')


@istest
class CompappTest1(QuickappTest):

    def compapp_test1(self):
        cases = []

        def add(args, ret):
            cases.append(dict(args=args, ret=ret))

        add('--compress --contracts -c clean;make --param1 10 --param2 1', 0)

        # parse error
        add('--compress --contracts -c clean;make  --param1 10 --parm2 1 ',
            QUICKAPP_USER_ERROR)

        # computation exception
        add('--compress --contracts  -c clean;make  --param1 10 --param2 -1 ',
            QUICKAPP_COMPUTATION_ERROR)

        for c in cases:
            args = c['args']

            if isinstance(args, str):
                args = args.split()
            tmpdir = mkdtemp()
            args = ['-o', tmpdir] + args
            ret = c['ret']
            with open(os.path.join(tmpdir, '.compmake.rc'), 'w') as f:
                f.write('config echo 1\n')
            ret_found = quickapp_main(QuickAppDemo2, args, sys_exit=False)
            msg = 'Expected %d, got %d.\nArguments: %s ' % (ret, ret_found, c['args'])
            self.assertEqual(ret, ret_found, msg)
            rmtree(tmpdir)


if __name__ == '__main__':
    quickapp_main(QuickAppDemo2)



import os
import time
from pprint import pformat

import numpy as np

from compmake import Context, Promise
from reprep.report_utils import StoreResults
from zuper_commons.ui import duration_compact
from conf_tools.utils import friendly_path
from zuper_commons.types import check_isinstance, describe_type, describe_value
from reprep import Report

from reprep.utils import frozendict2
from zuper_commons.text import natsorted
from . import logger
from .rm import write_report_single

__all__ = [
    'ReportManager',
]


class ReportManager:

    def __init__(self, context, outdir, index_filename=None):
        # TODO: remove context
        self.context = context
        self.outdir = outdir
        if index_filename is None:
            index_filename = os.path.join(self.outdir, 'report_index.html')
        self.index_filename = index_filename
        from reprep.report_utils import StoreResults
        self.allreports = StoreResults()
        self.allreports_filename = StoreResults()

        # report_type -> set of keys necessary
        self._report_types_format = {}

        self.html_resources_prefix = ''

        # check if we are called more than once; would be a bug
        self.index_job_created = False

        self.static_dir = os.path.join(self.outdir, 'reprep-static')

    def merge(self, other: "ReportManager") -> None:
        assert isinstance(other, ReportManager)
        """ Merges into this scructure the data from another reportmanager. """
        # print('I have: %d reports' % len(self.allreports))
        # print('merging with another with %d reports' % len(other.allreports))

        for key in other.allreports:
            report = other.allreports[key]

            if key in self.allreports:
                selfreport = self.allreports[key]
                msg = 'Found duplicate report %r' % key
                msg += ' jobs %s and %s' % (report, selfreport)
                # print(msg)
                if report.job_id != selfreport.job_id:
                    raise ValueError(msg)

            filename = other.allreports_filename[key]
            self.allreports[key] = report
            self.allreports_filename[key] = filename

    def set_html_resources_prefix(self, prefix):
        """
            Sets the prefix for the resources filename.

            example: set_resources_prefix('jbds')
        """
        self.html_resources_prefix = prefix + '-'

    def _check_report_format(self, report_type, **kwargs):
        keys = sorted(list(kwargs.keys()))
        # print('report %r %r' % (report_type, keys))
        if not report_type in self._report_types_format:
            self._report_types_format[report_type] = keys
        else:
            keys0 = self._report_types_format[report_type]
            if not keys == keys0:
                msg = 'Report %r %r' % (report_type, keys)
                msg += '\ndoes not match previous format %r' % keys0
                raise ValueError(msg)

    def get(self, report_type, **kwargs):
        key = frozendict2(report=report_type, **kwargs)
        return self.allreports[key]

    def add(self, context, report, report_type: str, **kwargs):
        """
            Adds a report to the collection.

            :param report: Promise of a Report object
            :param report_type: A string that describes the "type" of the report
            :param kwargs:  str->str,int,float  parameters used for grouping
        """
        from quickapp.compmake_context import CompmakeContext
        assert isinstance(context, CompmakeContext)
        if not isinstance(report_type, str):
            msg = 'Need a string for report_type, got %r.' % describe_value(report_type)
            raise ValueError(msg)

        if not isinstance(report, Promise):
            msg = ('ReportManager is mean to be given Promise objects, '
                   'which are the output of comp(). Obtained: %s'
                   % describe_type(report))
            raise ValueError(msg)

        # check the format is ok
        self._check_report_format(report_type, **kwargs)

        key = frozendict2(report=report_type, **kwargs)

        if key in self.allreports:
            msg = 'Already added report for %s' % key
            msg += '\n its values is %s' % self.allreports[key]
            msg += '\n new value would be %s' % report
            raise ValueError(msg)

        self.allreports[key] = report

        report_type_sane = report_type.replace('_', '')

        key_no_report = dict(**key)
        del key_no_report['report']
        basename = self.html_resources_prefix + report_type_sane
        if key_no_report:
            basename += '-' + basename_from_key(key_no_report)

        dirname = os.path.join(self.outdir, report_type_sane)
        filename = os.path.join(dirname, basename)
        self.allreports_filename[key] = filename + '.html'

        write_singles = False

        if write_singles:
            is_root = context.currently_executing == ['root']
            if not is_root:
                # Add also a single report independent of a global index

                # don't create the single report for the ones that are
                # defined in the root session

                filename_single = os.path.join(dirname, basename) + '_s.html'
                # filename_index_dyn = os.path.join(dirname, basename) + '_dyn.html'

                report_nid = self.html_resources_prefix + report_type_sane
                if key:
                    report_nid += '-' + basename_from_key(key)
                write_job_id = jobid_minus_prefix(context, report.job_id + '-writes')

                #                 write_report_yaml(report_nid, report_job_id=report.job_id,
                #                                   key=key, html_filename=filename_single,
                #                                   report_html_indexed=filename_index_dyn)

                context.comp(write_report_single,
                             report=report, report_nid=report_nid,
                             report_html=filename_single,
                             write_pickle=False,
                             static_dir=self.static_dir,
                             job_id=write_job_id)

    # @contract(context=Context)
    def create_index_job(self, context: Context):
        if self.index_job_created:
            msg = 'create_index_job() was already called once'
            raise ValueError(msg)
        self.index_job_created = True

        if not self.allreports:
            # no report necessary
            return

        create_write_jobs(context=context,
                          allreports_filename=self.allreports_filename,
                          allreports=self.allreports,
                          html_resources_prefix=self.html_resources_prefix,
                          index_filename=self.index_filename,
                          static_dir=self.static_dir,
                          suffix='write')


def create_write_jobs(context, allreports_filename, allreports,
                      html_resources_prefix, index_filename, suffix,
                      static_dir):
    # Do not pass as argument, it will take lots of memory!
    # XXX FIXME: there should be a way to make this update or not
    # otherwise new report do not appear
    optimize_space = False
    if optimize_space and len(allreports_filename) > 100:
        allreports_filename = context.comp_store(allreports_filename, 'allfilenames')
    else:
        allreports_filename = allreports_filename

    type2reports = sort_by_type(allreports_filename)

    for key in allreports:
        job_report = allreports[key]
        filename = allreports_filename[key]

        write_job_id = jobid_minus_prefix(context, job_report.job_id + '-' + suffix)

        # Create the links to report of the same type
        report_type = key['report']
        other_reports_same_type = type2reports[report_type]

        # find the closest report for different type
        others = find_others(type2reports, key)

        report_type_sane = report_type.replace('_', '')
        report_nid = html_resources_prefix + report_type_sane
        if key:
            report_nid += '-' + basename_from_key(key)

        key = dict(**key)
        del key['report']

        # XXX: not sure why this was here in the first place

        context.comp(write_report_and_update,
                     report=job_report, report_nid=report_nid,
                     report_html=filename, all_reports=allreports_filename,
                     index_filename=index_filename,
                     write_pickle=False,
                     this_report=key,
                     static_dir=static_dir,
                     other_reports_same_type=other_reports_same_type,
                     most_similar_other_type=others,
                     job_id=write_job_id)


def jobid_minus_prefix(context, want):
    prefix = context.get_comp_prefix()
    if prefix is not None:
        pref = prefix + '-'
        if want.startswith(pref):
            res = want[len(pref):]
        else:
            res = want
    else:
        res = want
    return res


def sort_by_type(allreports_filename):
    type2reports = {}
    for report_type, xs in allreports_filename.groups_by_field_value('report'):
        fields = xs.remove_field('report')
        # print(fields)
        from reprep.report_utils import StoreResults
        res = StoreResults()
        for k, v in list(fields.items()):
            res[k] = v
        type2reports[report_type] = res
    return type2reports


def find_others(type2reports, key):
    """ find the closest report for different type """
    report_type = key['report']

    key = dict(**key)
    del key['report']

    others = []
    for other_type, other_type_reports in list(type2reports.items()):
        if other_type == report_type:
            continue
        best = get_most_similar(other_type_reports, key)
        if best is not None:
            others.append((other_type, best, other_type_reports[best]))

    return others


def get_most_similar(reports_different_type, key):
    """ Returns the report of another type that is most similar to this report. """

    def score(key1):
        v1 = set(key1.values())
        v2 = set(key.values())
        return len(v1 & v2)

    keys = list(reports_different_type.keys())
    scores = np.array(list(map(score, keys)))

    tie = np.sum(scores == np.max(scores)) > 1
    if tie:
        # print('there is a tie: %s,\n %s' % (key, keys))
        return None

    i = int(np.argmax(scores))
    best = keys[i]

    return best


def create_links_html(this_report, other_reports_same_type, index_filename,
                      most_similar_other_type):
    check_isinstance(other_reports_same_type, StoreResults)
    '''
    :param this_report: dictionary with the keys describing the report
    :param other_reports_same_type: StoreResults -> filename
    :returns: html string describing the link
    '''

    def rel_link(f):  # (this is FROM f0 to f) --- trust me, it's ok
        f0 = other_reports_same_type[this_report]
        rl = os.path.relpath(f, os.path.dirname(f0))
        return rl

    s = ""

    # create table by cols
    table = create_links_html_table(this_report, other_reports_same_type)

    s += "<p><a href='%s'>All report</a></p>" % rel_link(index_filename)

    s += "<table class='variations'>"
    s += "<thead><tr>"
    for field, _ in table:
        s += "<th>%s</th>" % field
    s += "</tr></thead>"

    s += "<tr>"

    add_invalid_links = True

    for field, variations in table:
        s += "<td>"

        MAX_VARIATIONS_EXPLICIT = 10

        # hide the n/a
        # variations = [v for v in variations if v[1] is not None]

        all_vars = dict(variations)
        sorted_variations = natsorted([v[0] for v in variations])
        # print('sorted: %s' % sorted_variations)
        variations = [(v, all_vars[v]) for v in sorted_variations]

        if len(variations) > MAX_VARIATIONS_EXPLICIT:
            id_select = 'select-%s' % (field)
            onchange = 'onchange_%s' % (field)
            s += "<select id='%s' onChange='%s()'>\n" % (id_select, onchange)

            for text, link in variations:
                if link is not None:
                    s += "<option value='%s'>%s</a> \n" % (link, text)
                else:
                    if add_invalid_links:
                        s += "<option value=''>%s</a> \n" % (text)
                s += '<br/>'

            s += '</select>\n'

            s += """         
<script>
    $(function(){
      // bind change event to select
      $('#%s').bind('change', function () {
          var url = $(this).val(); // get selected value
          if (url) { // require a URL
              window.location = url; // redirect
          }
          return false;
      });
    });
</script>
""" % id_select

        else:
            for text, link in variations:
                if link is not None:
                    s += "<a href='%s'> %s</a> " % (link, text)
                else:
                    if add_invalid_links:
                        s += "%s " % (text)
                s += '<br/>\n'

        s += "</td>"

    s += "</tr>"
    s += "</table>"

    #     s += '<dl>'
    #     for other_type, most_similar, filename in most_similar_other_type:
    #         s += '<dt><a href="%s">%s</a></dt><dd>%s</dd>' % (rel_link(filename),
    #                                                           other_type, most_similar)
    #     s += '</dl>'

    if most_similar_other_type:
        s += '<p>Other report: '
        for other_type, _, filename in most_similar_other_type:
            s += '<a href="%s">%s</a> ' % (rel_link(filename), other_type)
        s += '</p>'

    s = '<div style="margin-left: 1em;">' + s + '</div>'
    return s


# @contract(returns="list( tuple(str, *))", other_reports_same_type=StoreResults)
def create_links_html_table(this_report, other_reports_same_type):
    # Iterate over all keys (each key gets a column)

    def rel_link(f):
        # TODO: make it relative
        f0 = other_reports_same_type[this_report]
        rl = os.path.relpath(f, os.path.dirname(f0))
        return rl

    cols = []
    fieldnames = other_reports_same_type.field_names()
    for field in fieldnames:
        field_values = other_reports_same_type.field_values(field)
        field_values = sorted(list(set(field_values)))
        col = []
        for fv in field_values:
            if fv == this_report[field]:
                # res = ('<span style="font-weight:bold">%s</span>' % str(fv), None)
                res = (str(fv), None)
            else:
                # this is the variation obtained by changing only one field value
                variation = dict(**this_report)
                variation[field] = fv
                # if it doesn't exist:
                if not variation in other_reports_same_type:
                    res = ('%s (n/a)' % str(fv), None)
                else:
                    res = (fv, rel_link(other_reports_same_type[variation]))
            col.append(res)
        cols.append((field, col))
    return cols


# @contract(report=Report, report_nid='str', other_reports_same_type=StoreResults)
def write_report_and_update(report, report_nid, report_html, all_reports, index_filename,
                            this_report,
                            other_reports_same_type,
                            most_similar_other_type,
                            static_dir,
                            write_pickle=False):
    if not isinstance(report, Report):
        msg = 'Expected Report, got %s.' % describe_type(report)
        raise ValueError(msg)

    links = create_links_html(this_report, other_reports_same_type, index_filename,
                              most_similar_other_type=most_similar_other_type)

    tree_html = '<pre style="display:none">%s</pre>' % report.format_tree()

    extras = dict(extra_html_body_start=links,
                  extra_html_body_end=tree_html)

    report.nid = report_nid
    html = write_report(report=report,
                        report_html=report_html,
                        static_dir=static_dir,
                        write_pickle=write_pickle, **extras)
    index_reports(reports=all_reports, index=index_filename, update=html)


# @contract(report=Report, report_html='str')
def write_report(report, report_html, static_dir, write_pickle=False, **kwargs):
    logger.debug('Writing to %s ' % friendly_path(report_html))
    #     if False:
    #         # Note here they might overwrite each other
    #         rd = os.path.join(os.path.dirname(report_html), 'images')
    #     else:
    rd = os.path.splitext(report_html)[0]
    report.to_html(report_html,
                   write_pickle=write_pickle,
                   resources_dir=rd,
                   static_dir=static_dir,
                   **kwargs)

    # TODO: save hdf format
    return report_html


# @contract(reports=StoreResults, index=str)
def index_reports(reports, index, update=None):  # @UnusedVariable
    """
        Writes an index for the report to the file given.
        The special key "report" gives the report type.

        report[dict(report=...,param1=..., param2=...) ] => filename
    """
    # print('Updating because of new report %s' % update)

    dirname = os.path.dirname(index)
    if not os.path.exists(dirname):
        os.makedirs(dirname)

    # logger.info('Writing on %s' % friendly_path(index))

    f = open(index, 'w')

    f.write("""
        <html>
        <head>
        <style type="text/css">
        span.when { float: right; }
        li { clear: both; }
        a.self { color: black; text-decoration: none; }
        </style>
        </head>
        <body>
    """)

    mtime = lambda x: os.path.getmtime(x)
    existing = list([x for x in list(reports.items()) if os.path.exists(x[1])])

    # create order statistics
    alltimes = np.array([mtime(b) for _, b in existing])

    def order(filename):
        """ returns between 0 and 1 the order statistics """
        assert os.path.exists(filename)
        histime = mtime(filename)
        compare = (alltimes < histime)
        return np.mean(compare * 1.0)

    def style_order(order):
        if order > 0.95:
            return "color: green;"
        if order > 0.9:
            return "color: orange;"
        if order < 0.5:
            return "color: gray;"
        return ""

    # @contract(k=dict, filename=str)
    def write_li(k, filename: str, element='li'):
        desc = ",  ".join('%s = %s' % (a, b) for a, b in list(k.items()))
        href = os.path.relpath(os.path.realpath(filename),
                               os.path.dirname(os.path.realpath(index)))
        if os.path.exists(filename):
            when = duration_compact(time.time() - mtime(filename))
            span_when = '<span class="when">%s ago</span>' % when
            style = style_order(order(filename))
            a = '<a href="%s">%s</a>' % (href, desc)
        else:
            # print('File %s does not exist yet' % filename)
            style = ""
            span_when = '<span class="when">missing</span>'
            a = '<a href="%s">%s</a>' % (href, desc)
        f.write('<%s style="%s">%s %s</%s>' % (element, style, a, span_when,
                                               element))

    # write the first 10
    existing.sort(key=lambda x: (-mtime(x[1])))
    nlast = min(len(existing), 10)
    last = existing[:nlast]
    f.write('<h2 id="last">Last %d report</h2>\n' % (nlast))

    f.write('<ul>')
    for i in range(nlast):
        write_li(*last[i])
    f.write('</ul>')

    if False:
        for report_type, r in reports.groups_by_field_value('report'):
            f.write('<h2 id="%s">%s</h2>\n' % (report_type, report_type))
            f.write('<ul>')
            r = reports.select(report=report_type)
            items = list(r.items())
            items.sort(key=lambda x: str(x[0]))  # XXX use natsort
            for k, filename in items:
                write_li(k, filename)

            f.write('</ul>')

    f.write('<h2>All report</h2>\n')

    try:
        sections = make_sections(reports)
    except:
        logger.error(str(list(reports.keys())))
        raise

    if sections['type'] == 'sample':
        # only one...
        sections = dict(type='division', field='raw',
                        division=dict(raw1=sections), common=dict())

    def write_sections(sections, parents):
        assert 'type' in sections
        assert sections['type'] == 'division', sections
        field = sections['field']
        division = sections['division']

        f.write('<ul>')
        sorted_values = natsorted(list(division.keys()))
        for value in sorted_values:
            parents.append(value)
            html_id = "-".join(map(str, parents))
            bottom = division[value]
            if bottom['type'] == 'sample':
                d = {field: value}
                if not bottom['key']:
                    write_li(k=d, filename=bottom['value'], element='li')
                else:
                    f.write('<li> <p id="%s"><a class="self" href="#%s">%s = %s</a></p>\n'
                            % (html_id, html_id, field, value))
                    f.write('<ul>')
                    write_li(k=bottom['key'], filename=bottom['value'], element='li')
                    f.write('</ul>')
                    f.write('</li>')
            else:
                f.write('<li> <p id="%s"><a class="self" href="#%s">%s = %s</a></p>\n'
                        % (html_id, html_id, field, value))

                write_sections(bottom, parents)
                f.write('</li>')
        f.write('</ul>')

    write_sections(sections, parents=[])

    f.write('''
    
    </body>
    </html>
    
    ''')
    f.close()


def make_sections(allruns, common=None):
    # print allruns.keys()
    if common is None:
        common = {}

    # print('Selecting %d with %s' % (len(allruns), common))

    if len(allruns) == 1:
        key = list(allruns.keys())[0]
        value = allruns[key]
        return dict(type='sample', common=common, key=key, value=value)

    fields_size = [(field, len(list(allruns.groups_by_field_value(field))))
                   for field in allruns.field_names_in_all_keys()]

    # Now choose the one with the least choices
    fields_size.sort(key=lambda x: x[1])

    if not fields_size:
        # [frozendict({'i': 1, 'n': 3}), frozendict({'i': 2, 'n': 3}), frozendict({}), frozendict({'i': 0, 'n': 3})]
        msg = 'Not all records of the same type have the same fields'
        msg += pformat(list(allruns.keys()))
        raise ValueError(msg)

    field = fields_size[0][0]
    division = {}
    for value, samples in allruns.groups_by_field_value(field):
        samples = samples.remove_field(field)
        c = dict(common)
        c[field] = value
        try:
            division[value] = make_sections(samples, common=c)
        except Exception:
            msg = 'Error occurred inside grouping by field %r = %r' % (field, value)
            msg += '\nCommon: %r' % common
            msg += '\nSamples: %s' % list(samples.keys())
            logger.error(msg)
            raise

    return dict(type='division', field=field,
                division=division, common=common)


def _dynreports_create_index(context, merged_data):
    rm = merged_data['report_manager']
    rm.create_index_job(context)


def basename_from_key(key: dict) -> str:
    """ Returns a nice basename from a key
        that doesn't have special chars """
    if not key:
        raise ValueError('empty key')
    keys_ordered = sorted(list(key.keys()))
    values = []
    for k in keys_ordered:
        value = key[k]
        value = str(value)
        value = value.replace('_', '')
        value = value.replace('-', '')
        value = value.replace('.', '')
        values.append(value)
    basename = "-".join(values)
    basename = basename.replace('/', '_')  # XXX
    return basename



from __future__ import print_function
from copy import deepcopy

from contracts import contract

from compmake import Promise
from quickapp import CompmakeContext
from quickapp.report_manager import basename_from_key
from reprep import NotExistent, Report, logger


__all__ = ['ReportProxy', 'get_node']


class FigureProxy(object):
    def __init__(self, id_figure, report_proxy):
        self.id_figure = id_figure
        self.report_proxy = report_proxy
        
    @contract(what='str')
    def sub(self, what, **kwargs):
        self.report_proxy.op(rp_figure_sub, id_figure=self.id_figure, what=what, **kwargs)
    
    
class ReportProxy(object):
    
    @contract(context=CompmakeContext)
    def __init__(self, context):
        self.context = context
        self.operations = []
        self.resources = {}
        
    def op(self, function, **kwargs):
        self.operations.append((function, kwargs))
    
    @contract(nid='str')
    def figure(self, nid, **kwargs):
        if nid is None:
            nid = 'Figure'
        self.op(rp_create_figure, id_parent='report', nid=nid, **kwargs)
        return FigureProxy(nid, self)
    
    @contract(child=Promise, nid='str')
    def add_child_with_id(self, child, nid):
        self.op(add_child_with_id, id_parent='report', child=child, nid=nid)
    
    def add_child_from_other(self, url, nid, report_type, strict=True, **report_args):
        part = self.get_part_of(url, report_type, strict=strict, ** report_args)
        
        if nid is None:
            nid = basename_from_key(report_args) + '-' + url.replace('/', '-')  # XXX url chars
            
        self.add_child_with_id(part, nid)
        return nid
 
    @contract(returns=Promise, url=str, report_type=str)
    def get_part_of(self, url, report_type, strict=True, **report_args):
        job_id = 'get_part-' + report_type + '-' + basename_from_key(report_args)
        r = self.context.get_report(report_type, **report_args)
        job_id += '-' + url.replace('/', '_')  # XXX
        part = self.context.comp(get_node, url=url, r=r, strict=strict, job_id=job_id)
        return part
    
    @contract(returns=Promise)
    def get_job(self):
        return self.context.comp(execute_proxy, self.operations) 
    

@contract(url=str, r=Report, returns=Report)
def get_node(url, r, strict=True):
    try:
        node = r.resolve_url(url)
    except NotExistent as e:
        if strict:
            logger.error('Error while getting url %r\n%s' % (url,
                                                             r.format_tree()))
            raise
        else:
            logger.warn('Ignoring error: %s' % e)
            return Report()

    node = deepcopy(node)
    node.parent = None
    return node


@contract(resources='dict', id_parent='str', child=Report, nid='str')
def add_child_with_id(resources, id_parent, child, nid):
    parent = resources[id_parent]
    child.nid = nid
    print(child.format_tree())
    parent.add_child(child)

    
@contract(resources='dict', id_parent='str', nid='str')
def rp_create_figure(resources, id_parent, nid, **figargs):
    parent = resources[id_parent]
    resources[nid] = parent.figure(nid=nid, **figargs)


@contract(resources='dict', id_figure='str', what='str')
def rp_figure_sub(resources, id_figure, what, caption=None):
    figure = resources[id_figure]
    try:
        figure.sub(what, caption=caption)
    except (NotExistent, Exception) as e:
        logger.error(e)
        

def execute_proxy(operations):
    report = Report()
    resources = {}
    resources['report'] = report
    for what, kwargs in operations:
        what(resources=resources, **kwargs)
    print(report.format_tree())
    return report



from reprep import Report

__all__ = [
    'write_report_single',
]

from zuper_commons.types import check_isinstance


def write_report_single(report: Report,
                        report_nid, report_html,
                        static_dir,
                        write_pickle=False):
    from quickapp.report_manager import write_report

    check_isinstance(report, Report)
    report.nid = report_nid
    write_report(report, report_html, static_dir=static_dir, write_pickle=write_pickle)
#
#
# def write_report_yaml(report_nid, report_job_id, key, html_filename, report_html_indexed):
#
#     metadata_file = os.path.splitext(html_filename)[0] + '.rm_reports.yaml'
#     rel_filename = os.path.relpath(os.path.realpath(report_html_indexed),
#                                    os.path.dirname(os.path.realpath(metadata_file)))
#
#     entry = dict(id=report_nid, desc='Automatically generated report',
#                  code=['quickapp.rm.GeneratedReport',
#                        {'key': dict(**key),
#                         "file:filename": rel_filename,
#                         'report_job_id': report_job_id}])
#     dirname = os.path.dirname(metadata_file)
#     if not os.path.exists(dirname):
#         try:
#             os.makedirs(dirname)
#         except:
#             pass  # XXX: race condition
#     with open(metadata_file, 'w') as f:
#         f.write(yaml.dump([entry], default_flow_style=False))
#
# #
# def create_job_index_dynamic(context, dirname, index_filename, html_resources_prefix, static_dir):
#     """ Load the dynamically-generated report """
#     if not os.path.exists(dirname):
#         print('Reports directory not found. You should rerun this job later.')
#         return
#
#     reports = get_conftools_rm_reports()
#
#     print(reports.dirs_read)
#     reports.force_load(dirname)
#
#     id_reports = list(reports.keys())
#     if not id_reports:
#         print('No report found yet.')
#         return
#
#     allreports = StoreResults()
#     allreports_filename = StoreResults()
#     for id_report in id_reports:
#         report = reports.instance(id_report)
#         filename = report.filename
#         key = report.key
#         report_job_id = report.report_job_id
#         allreports_filename[key] = filename
#         db = context.get_compmake_db()
#         if job_exists(report_job_id, db=db):
#             allreports[key] = Promise(report_job_id)
#         else:
#             print('Warning: yaml found for job which is not there yet.')
#             print('     job: %s' % report_job_id)
#
#     from quickapp.report_manager import create_write_jobs
#     create_write_jobs(context,
#                       allreports_filename,
#                       allreports,
#                       html_resources_prefix,
#                       index_filename,
#                       suffix='writedyn',
#                       static_dir=static_dir)



import os

from setuptools import find_packages, setup


def get_version(filename):
    import ast
    version = None
    with open(filename) as f:
        for line in f:
            if line.startswith('__version__'):
                version = ast.parse(line).body[0].value.s
                break
        else:
            raise ValueError('No version found in %r.' % filename)
    if version is None:
        raise ValueError(filename)
    return version


version = get_version('src/quickapp/__init__.py')

description = """"""


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


long_description = read('README.rst')

line = 'z6'

install_requires = [
    'compmake-z6',
    'reprep-z6',
    'PyContracts3',
    'DecentLogs-z6',
    'ConfTools-z6',
]

setup(name=f'QuickApp-{line}',
      author="Andrea Censi",
      author_email="",
      url='http://github.com/AndreaCensi/quickapp',

      description=description,
      long_description=long_description,
      keywords="",
      license="",

      classifiers=[
          'Development Status :: 4 - Beta',
          # 'Intended Audience :: Developers',
          # 'License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)',
          # 'Topic :: Software Development :: Quality Assurance',
          # 'Topic :: Software Development :: Documentation',
          # 'Topic :: Software Development :: Testing'
      ],

      version=version,
      download_url='http://github.com/AndreaCensi/quickapp/tarball/%s' % version,

      entry_points={
          'console_scripts': [
              # 'comptests = comptests:main_comptests'
          ]
      },
      package_dir={'': 'src'},
      packages=find_packages('src'),
      install_requires=install_requires,
      tests_require=['nose'],
      )



# 
# 
# __all__ = ['GeneratedReport']
# 
# 
# class GeneratedReport():
# 
#     def __init__(self, filename, key, report_job_id):
#         self.filename = filename
#         self.key = key
#         self.report_job_id = report_job_id



#!/usr/bin/env python
from nose.tools import istest

from quickapp import QuickApp, iterate_context_names
from reprep import Report
from quickapp.tests.quickappbase import QuickappTest



def report_example1(param1, param2):
    r = Report()
    r.text('type', 'This is one report')
    r.text('param1', '%s' % param1)
    r.text('param2', '%s' % param2)
    return r

def report_example2(param1, param2):
    r = Report()
    r.text('type', 'This is another report')
    r.text('param1', '%s' % param1)
    r.text('param2', '%s' % param2)
    return r

def instance_reports1(context):
    param1s = ['a', 'b']
    for c1, param1 in iterate_context_names(context, param1s, key='param1'):
        c1.comp_dynamic(instance_reports2, param1=param1)

def instance_reports2(context, param1):
    param2s = [1, 2]
    for c2, param2 in iterate_context_names(context, param2s, key='param2'):
        c2.comp_dynamic(instance_reports3, param1=param1, param2=param2)   
     
def instance_reports3(context, param1, param2):
    context.comp(dummy)

    r = context.comp(report_example2, param1=param1, param2=param2)
    context.add_report(r, 'report_example2')

    context.comp_dynamic(instance_reports4, param1=param1, param2=param2)
    
def instance_reports4(context, param1, param2):
    r = context.comp(report_example1, param1=param1, param2=param2)
    context.add_report(r, 'report_example1')
    

def dummy():
    pass

class QuickAppDemoReport2(QuickApp):

    def define_options(self, params):
        pass

    def define_jobs_context(self, context):
        context.comp_dynamic(instance_reports1) 

@istest
class CompappTest1(QuickappTest):

    def compapp_test1(self):
        self.run_quickapp(QuickAppDemoReport2, cmd='make recurse=1') 


if __name__ == '__main__':
    main = QuickAppDemoReport2.get_sys_main()
    main()



#!/usr/bin/env python
# from quickapp import QuickApp
# from reprep import Report
# import sys
# 
# def actual_computation(param1, param2, iteration):
#     pass
# 
# def report(param2, jobs):  # @UnusedVariable
#     r = Report()
#     
#     return r
# 
# class QuickAppDemo1(QuickApp):
#     
#     def define_options(self, params):
#         params.add_int('param1', help='First parameter', default=1)
#         params.add_int_list('param2', help='Second parameter')
#         
#     def define_jobs_context(self, context):
#         options = self.get_options()
#         
#         jobs = self.comp_comb(actual_computation,
#                               param1=options.param1,
#                               param2=options.param2,
#                               iteration=QuickApp.choice(range(4)))
#         
#         rm = self.get_report_manager()
#         for param2, samples in jobs.groups_by_field_value('param2'):
#             rj = self.comp(report, param2, samples)
#             rm.add(rj, 'report', param2=param2)
# 
# 
#         
# def compapp_test1():
#     args = ['-o', 'quickapp_test1',
#             '-c', 'make all', '--param1', '10', '--param2', '1,2,3']
#     QuickAppDemo1().main(args)
#     
# 
# if __name__ == '__main__':
#     sys.exit(QuickAppDemo1().main())
#     



#!/usr/bin/env python

from quickapp import QuickApp


class AppExample(QuickApp):
    """ Simplest app example """

    def define_options(self, params):
        params.add_int('x', default=1)

    def define_jobs_context(self, context):
        options = self.get_options()
        # create a job
        context.comp(f, options.x)

def f(x):
    print('x = %s' % x)        

app_example_main = AppExample.get_sys_main()

if __name__ == '__main__':
    app_example_main()